from copy import deepcopy
from dataclasses import dataclass
import os
from typing import Collection, Optional, Tuple
from unittest.mock import ANY, call, Mock, patch

import click
import pytest

from anyscale.anyscale_pydantic import ValidationError
from anyscale.client.openapi_client.models.decoratedlistserviceapimodel_list_response import (
    DecoratedlistserviceapimodelListResponse,
)
from anyscale.client.openapi_client.models.list_response_metadata import (
    ListResponseMetadata,
)
from anyscale.client.openapi_client.models.rollback_service_model import (
    RollbackServiceModel,
)
from anyscale.controllers.service_controller import ServiceController
from anyscale.models.service_model import ServiceConfig
from anyscale.sdk.anyscale_client.configuration import Configuration
from anyscale.sdk.anyscale_client.models import (
    ApplyServiceModel,
    Cluster,
    ClusterResponse,
    ServiceModel,
    ServicemodelResponse,
)
from anyscale.utils.workload_types import Workload


TEST_SERVICE_ID = "test-service-id"
TEST_SERVICE_NAME = "test-service-name"

OPENAPI_NO_VALIDATION = Configuration()
OPENAPI_NO_VALIDATION.client_side_validation = False


class FakeServiceSDK:
    def __init__(self):
        self.reset()

    def reset(self):
        self.rolled_out_service_model: Optional[ApplyServiceModel] = None
        self.rolled_back_service_id: Optional[str] = None
        self.rolled_back_service_model: Optional[RollbackServiceModel] = None
        self.terminated_service_id: Optional[str] = None

    def _service_response_model(self, service_id: str) -> ServicemodelResponse:
        return ServicemodelResponse(
            result=ServiceModel(
                id=service_id, local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

    def rollout_service(
        self, apply_service_model: ApplyServiceModel
    ) -> ServicemodelResponse:
        self.rolled_out_service_model = apply_service_model
        return self._service_response_model(TEST_SERVICE_ID)

    def rollback_service(
        self, service_id: str, rollback_service_model: RollbackServiceModel
    ) -> ServicemodelResponse:
        self.rolled_back_service_id = service_id
        self.rolled_back_service_model = rollback_service_model
        return self._service_response_model(service_id)

    def terminate_service(self, service_id: str) -> ServicemodelResponse:
        self.terminated_service_id = service_id
        return self._service_response_model(service_id)

    def get_cluster(self, session_id: str) -> ClusterResponse:
        return ClusterResponse(
            result=Cluster(
                id=session_id, local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )


@pytest.fixture()
def setup_service_controller(
    mock_auth_api_client,
) -> Tuple[ServiceController, FakeServiceSDK]:
    fake_sdk = FakeServiceSDK()
    return ServiceController(sdk=fake_sdk), fake_sdk


def test_service_config_model_with_entrypoint():
    config_dict = {
        "name": "test_service",
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.ray_serve_config is None


def test_service_config_model_with_ray_serve_config():
    runtime_env = {
        "pip": ["requests"],
        "working_dir": "https://github.com/anyscale/docs_examples/archive/refs/heads/main.zip",
    }
    config_dict = {
        "name": "test_service",
        "ray_serve_config": {
            "import_path": "serve_hello:entrypoint",
            "runtime_env": runtime_env,
        },
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.runtime_env is None
        assert service_config.ray_serve_config["runtime_env"] == runtime_env


def test_service_config_model_with_ray_serve_config_working_dir():
    runtime_env = {"pip": ["requests"], "working_dir": "."}
    config_dict = {
        "name": "test_service",
        "ray_serve_config": {
            "import_path": "serve_hello:entrypoint",
            "runtime_env": runtime_env,
        },
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.ray_serve_config["runtime_env"] == runtime_env
        assert service_config.ray_gcs_external_storage_config is None


def test_service_config_model_with_ray_gcs_external_storage_config():
    config_dict = {
        "name": "test_service",
        "ray_serve_config": {"runtime_env": {"pip": ["requests"], "working_dir": "."}},
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
        "ray_gcs_external_storage_config": {"address": "addr:port",},
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert (
            service_config.ray_gcs_external_storage_config
            == config_dict["ray_gcs_external_storage_config"]
        )
        assert service_config.runtime_env is None


class TestGenerateConfig:
    @pytest.mark.parametrize(
        ("ray_serve_config", "runtime_env"),
        [[{"runtime_env": {"pip": ["requests"], "working_dir": "."}}, None],],
    )
    def test_override_config_options(
        self,
        setup_service_controller,
        ray_serve_config: Optional[str],
        runtime_env: Optional[str],
    ):
        """
        This test checks that the config file for Service v2
        can be properly parsed.
        """
        service_controller, _ = setup_service_controller

        name = "test_service_name"
        description = "mock_description"
        build_id = "test_build_id"
        compute_config_id = "test_compute_config_id"
        project_id = "test_project_id"
        rollout_strategy = "ROLLOUT"
        canary_percent = 100
        config_dict = {
            "name": name,
            "description": description,
            "ray_serve_config": ray_serve_config,
            "runtime_env": runtime_env,
            "project_id": project_id,
            "build_id": build_id,
            "compute_config_id": compute_config_id,
            "version": "test_abc",
            "canary_percent": canary_percent,
            "rollout_strategy": rollout_strategy,
        }
        mock_validate_successful_build = Mock()
        with patch.multiple(
            "anyscale.models.job_model",
            validate_successful_build=mock_validate_successful_build,
        ):
            config = service_controller.override_config_options(
                ServiceConfig.parse_obj(config_dict), auto_complete_rollout=True
            )

            expected_service_config = ServiceConfig(
                name=name,
                description=description,
                ray_serve_config=ray_serve_config,
                runtime_env=runtime_env,
                project_id=project_id,
                build_id=build_id,
                compute_config_id=compute_config_id,
                version="test_abc",
                canary_percent=canary_percent,
                rollout_strategy=rollout_strategy,
                auto_complete_rollout=True,
            )
        assert config == expected_service_config

    @pytest.mark.parametrize("add_extra_field", [True, False])
    def test_config_validation(self, add_extra_field, setup_service_controller):
        """
        This test checks that an error is thrown if extra/mispelled fields
        are added into the config file
        """
        service_controller, _ = setup_service_controller

        name = "test_service_name"
        description = "mock_description"
        build_id = "test_build_id"
        compute_config_id = "test_compute_config_id"
        project_id = "test_project_id"
        rollout_strategy = "ROLLOUT"
        extra_field_value = "*****"
        canary_percent = 100
        config_dict = {
            "name": name,
            "description": description,
            "ray_serve_config": {},
            "project_id": project_id,
            "build_id": build_id,
            "compute_config_id": compute_config_id,
            "version": "test_abc",
            "canary_percent": canary_percent,
            "rollout_strategy": rollout_strategy,
        }

        if add_extra_field:
            extra_field_value = "*****"
            config_dict["extra_field"] = extra_field_value

        with patch.multiple(
            "anyscale.models.job_model", validate_successful_build=Mock(),
        ):
            if add_extra_field:
                with pytest.raises(ValidationError):
                    service_controller.parse_service_config_dict(config_dict)
            else:
                service_controller.parse_service_config_dict(config_dict)

    @pytest.mark.parametrize(
        (
            "config_rollout_strategy",
            "config_max_surge_percent",
            "input_rollout_strategy",
            "input_max_surge_percent",
            "should_error",
        ),
        [
            # --max-surge-percent is not supported for in-place updates.
            ("IN_PLACE", 10.0, None, None, True),
            (None, None, "IN_PLACE", 10.0, True),
            ("IN_PLACE", None, None, 10.0, True),
            (None, 10.0, "IN_PLACE", None, True),
            ("IN_PLACE", 10.0, "IN_PLACE", 10.0, True),
            # --max-surge-percent is supported for rolling updates.
            ("ROLLING", 10.0, None, None, False),
            (None, None, "ROLLING", 10.0, False),
            ("ROLLING", None, None, 10.0, False),
            (None, 10.0, "ROLLING", None, False),
            ("ROLLING", 10.0, "ROLLING", 10.0, False),
            (None, 10.0, None, 10.0, False),
        ],
    )
    def test_override_config_options_max_surge_percent_in_place(
        self,
        setup_service_controller,
        config_rollout_strategy,
        config_max_surge_percent,
        input_rollout_strategy,
        input_max_surge_percent,
        should_error,
    ):
        """Check validation for max_surge_percent and rollout_strategy."""
        service_controller, _ = setup_service_controller

        name = "test_service_name"
        description = "mock_description"
        ray_serve_config = {}
        build_id = "test_build_id"
        compute_config_id = "test_compute_config_id"
        project_id = "test_project_id"
        canary_percent = 100

        config_dict = {
            "name": name,
            "description": description,
            "ray_serve_config": ray_serve_config,
            "project_id": project_id,
            "build_id": build_id,
            "compute_config_id": compute_config_id,
            "version": "test_abc",
            "canary_percent": canary_percent,
            "rollout_strategy": config_rollout_strategy,
            "max_surge_percent": config_max_surge_percent,
        }

        mock_validate_successful_build = Mock()
        with patch.multiple(
            "anyscale.models.job_model",
            validate_successful_build=mock_validate_successful_build,
        ):

            if should_error:
                with pytest.raises(ValueError):
                    service_controller.override_config_options(
                        ServiceConfig.parse_obj(config_dict),
                        auto_complete_rollout=True,
                        rollout_strategy=input_rollout_strategy,
                        max_surge_percent=input_max_surge_percent,
                    )
            else:
                service_controller.override_config_options(
                    ServiceConfig.parse_obj(config_dict),
                    auto_complete_rollout=True,
                    rollout_strategy=input_rollout_strategy,
                    max_surge_percent=input_max_surge_percent,
                )


@pytest.mark.parametrize(
    ("working_dir"),
    [
        ("https://github.com/anyscale/docs_examples/archive/refs/heads/main.zip"),
        ("s3://already-a-uri/path"),
    ],
)
def test_generate_config_from_file_inside_workspace_no_working_dir_override(
    setup_service_controller, working_dir: str,
):
    """
    This test checks that if the service v2 is launched from the workspace with
    a remote working dir being specified, the method would inspect the remote url
    and not override it.
    """
    service_controller, _ = setup_service_controller

    name = "test_service_name"
    description = "mock_description"
    ray_serve_config = {
        "runtime_env": {"pip": ["requests"], "working_dir": working_dir,}
    }
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    canary_percent = 100
    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": canary_percent,
        "rollout_strategy": rollout_strategy,
    }
    mock_validate_successful_build = Mock()

    workspace_id = "Test_workspace_id"
    with patch.dict(
        os.environ,
        {
            "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": workspace_id,
            "ANYSCALE_SESSION_ID": "test_session_id",
        },
    ), patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        config = service_controller.parse_service_config_dict(config_dict)
        assert config == ServiceConfig(
            name=name,
            description=description,
            ray_serve_config=ray_serve_config,
            project_id=project_id,
            build_id=build_id,
            compute_config_id=compute_config_id,
            version="test_abc",
            canary_percent=canary_percent,
            rollout_strategy=rollout_strategy,
        )


@pytest.mark.parametrize("override_name", [None, "override_service_name"])
@pytest.mark.parametrize("override_version", [None, "override_version"])
@pytest.mark.parametrize(
    ("override_canary_percent"), [None, 0, 40],
)
@pytest.mark.parametrize("override_rollout_strategy", [None, "IN_PLACE"])
@pytest.mark.parametrize("auto_complete_rollout", [True, False])
@pytest.mark.parametrize("max_surge_percent", [None, 20])
def test_rollout(  # noqa: PLR0913
    setup_service_controller,
    override_name: Optional[str],
    override_version: Optional[str],
    override_canary_percent: Optional[int],
    override_rollout_strategy: Optional[str],
    auto_complete_rollout: bool,
    max_surge_percent: Optional[int],
) -> None:
    """
    This tests that when we submit a Service config with ray_serve_config,
    we submit both the Service v1 and Service v2 configs.
    Please reference services_internal_api_models.py for more details.
    """
    service_controller, fake_sdk = setup_service_controller

    name = "test_service_name"
    description = "mock_description"
    ray_serve_config = {"runtime_env": {"pip": ["requests"], "working_dir": "."}}
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    service_config = {"access": {"use_bearer_token": True}}

    if override_rollout_strategy == "IN_PLACE":
        # Cannot set max_surge_percent during an in-place rollout.
        max_surge_percent = None

    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": 100,
        "rollout_strategy": rollout_strategy,
        "config": service_config,
    }
    mock_validate_successful_build = Mock()
    mock_override_runtime_env_config = Mock(
        return_value={"pip": ["requests"], "working_dir": "."}
    )
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ), patch.multiple(
        "anyscale.controllers.service_controller",
        override_runtime_env_config=mock_override_runtime_env_config,
    ):
        service_controller.rollout(
            ServiceConfig.parse_obj(config_dict),
            name=override_name,
            version=override_version,
            canary_percent=override_canary_percent,
            rollout_strategy=override_rollout_strategy,
            auto_complete_rollout=auto_complete_rollout,
            max_surge_percent=max_surge_percent,
        )

    assert fake_sdk.rolled_out_service_model == ApplyServiceModel(
        name=override_name if override_name else name,
        description=description,
        build_id=build_id,
        compute_config_id=compute_config_id,
        project_id=project_id,
        ray_serve_config=ray_serve_config,
        version=override_version if override_version else "test_abc",
        canary_percent=override_canary_percent
        if override_canary_percent is not None
        else 100,
        rollout_strategy=override_rollout_strategy
        if override_rollout_strategy
        else "ROLLOUT",
        config=service_config,
        auto_complete_rollout=auto_complete_rollout,
        max_surge_percent=max_surge_percent,
    )


@dataclass
class FakeServiceResult:
    id: str
    name: str
    project_id: str = "fake-default-project-id"
    current_state: str = "RUNNING"


class TestGetServiceID:
    def test_invalid_inputs(self, setup_service_controller):
        service_controller, _ = setup_service_controller

        with pytest.raises(
            click.ClickException,
            match="Service ID, name, or config file must be specified.",
        ):
            service_controller.get_service_id()

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_id="test-id", service_name="test-name"
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_id="test-id",
                service_config_file="test-config-file-path",
                project_id=None,
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_id="test-id",
                service_name="test-name",
                service_config_file="test-config-file-path",
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_name="test-name", service_config_file="test-config-file-path"
            )

        with pytest.raises(click.ClickException, match="not found"):
            service_controller.get_service_id(
                service_config_file="/fake/service/config/path"
            )

    def test_service_id_provided(self, setup_service_controller):
        """If the ID is provided, the same ID should always be returned and no API calls made."""
        service_controller, _ = setup_service_controller

        assert service_controller.get_service_id(service_id="test-id") == "test-id"
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_not_called()

    def test_name_provided_no_matching_service(self, setup_service_controller):
        """If a name is provided but no services match it, an exception should be raised."""
        service_controller, _ = setup_service_controller

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[], metadata=Mock(next_paging_token=None),
        )

        with pytest.raises(
            click.ClickException,
            match=f"No service with name '{TEST_SERVICE_NAME}' was found",
        ):
            service_controller.get_service_id(service_name=TEST_SERVICE_NAME)

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id=None,
            paging_token=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        # Test the config file codepath.
        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller.read_service_config_file = Mock()
        service_controller.parse_service_config_dict = Mock(
            return_value=mock_service_config,
        )

        with pytest.raises(
            click.ClickException,
            match=f"No service with name '{TEST_SERVICE_NAME}' was found",
        ):
            service_controller.get_service_id(service_config_file="test-config-file")

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id=None,
            paging_token=None,
        )

    def test_name_provided_single_matching_service(self, setup_service_controller):
        """If a name is provided and one service matches it, its ID should be returned."""
        service_controller, _ = setup_service_controller

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[FakeServiceResult(id=TEST_SERVICE_ID, name=TEST_SERVICE_NAME)],
            metadata=Mock(next_paging_token=None),
        )

        assert (
            service_controller.get_service_id(service_name=TEST_SERVICE_NAME)
            == TEST_SERVICE_ID
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id=None,
            paging_token=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        # Test the config file codepath.
        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller.read_service_config_file = Mock()
        service_controller.parse_service_config_dict = Mock(
            return_value=mock_service_config,
        )
        assert (
            service_controller.get_service_id(service_config_file="test-config-file")
            == TEST_SERVICE_ID
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id=None,
            paging_token=None,
        )

    def test_name_provided_multiple_matching_services(self, setup_service_controller):
        """If a name is provided and multiple services match it, an exception should be raised."""
        service_controller, _ = setup_service_controller

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[
                FakeServiceResult(id="test-service-id-1", name=TEST_SERVICE_NAME),
                FakeServiceResult(id="test-service-id-2", name=TEST_SERVICE_NAME),
            ],
            metadata=Mock(next_paging_token=None),
        )

        with pytest.raises(
            click.ClickException,
            match=f"There are multiple services with name '{TEST_SERVICE_NAME}'.",
        ):
            service_controller.get_service_id(service_name=TEST_SERVICE_NAME)

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id=None,
            paging_token=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller.read_service_config_file = Mock()
        service_controller.parse_service_config_dict = Mock(
            return_value=mock_service_config,
        )

        with pytest.raises(
            click.ClickException,
            match=f"There are multiple services with name '{TEST_SERVICE_NAME}'.",
        ):
            service_controller.get_service_id(service_config_file="test-config-file")

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id=None,
            paging_token=None,
        )

    def test_name_provided_exact_match(self, setup_service_controller):
        """The API uses a substring match, so ensure we filter to the exact name."""
        service_controller, _ = setup_service_controller

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[
                FakeServiceResult(id="test-service-id-1", name="test-service"),
                FakeServiceResult(id="test-service-id-2", name="test-service-suffix"),
            ],
            metadata=Mock(next_paging_token=None),
        )

        assert (
            service_controller.get_service_id(service_name="test-service")
            == "test-service-id-1"
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name="test-service",
            project_id=None,
            paging_token=None,
        )

    def test_no_project_id(self, setup_service_controller):
        """If there is no project_id, the results should not be filtered based on project."""
        service_controller, _ = setup_service_controller

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[FakeServiceResult(id=TEST_SERVICE_ID, name=TEST_SERVICE_NAME)],
            metadata=Mock(next_paging_token=None),
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(service_name=TEST_SERVICE_NAME)
                == TEST_SERVICE_ID
            )
            service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
                count=10,
                creator_id=None,
                name=TEST_SERVICE_NAME,
                project_id=None,
                paging_token=None,
            )

    def test_project_id_in_service_config(self, setup_service_controller):
        """If the project_id is passed in the service config it should be used as a filter."""
        service_controller, _ = setup_service_controller

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[
                FakeServiceResult(
                    id=TEST_SERVICE_ID,
                    name=TEST_SERVICE_NAME,
                    project_id="test-project-id",
                )
            ],
            metadata=Mock(next_paging_token=None),
        )

        mock_service_config = Mock()
        mock_service_config.configure_mock(
            name=TEST_SERVICE_NAME, project_id="test-project-id"
        )
        service_controller.read_service_config_file = Mock()
        service_controller.parse_service_config_dict = Mock(
            return_value=mock_service_config,
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(
                    service_config_file="test-config-file"
                )
                == TEST_SERVICE_ID
            )
            service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
                count=10,
                creator_id=None,
                name=TEST_SERVICE_NAME,
                project_id="test-project-id",
                paging_token=None,
            )

    def test_override_project_id(self, setup_service_controller):
        """If the project_id is passed directly, the passed value should be used to filter."""
        service_controller, _ = setup_service_controller

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[
                FakeServiceResult(
                    id=TEST_SERVICE_ID,
                    name=TEST_SERVICE_NAME,
                    project_id="overridden-project-id",
                )
            ],
            metadata=Mock(next_paging_token=None),
        )

        # Test overriding when getting service ID by name.
        assert (
            service_controller.get_service_id(
                service_name=TEST_SERVICE_NAME, project_id="overridden-project-id"
            )
            == TEST_SERVICE_ID
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id="overridden-project-id",
            paging_token=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        # Test overriding when getting service ID by config file.
        mock_service_config = Mock()
        mock_service_config.configure_mock(
            name=TEST_SERVICE_NAME, project_id="test-project-id"
        )
        service_controller.read_service_config_file = Mock()
        service_controller.parse_service_config_dict = Mock(
            return_value=mock_service_config,
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(
                    service_config_file="test-config-file",
                    project_id="overridden-project-id",
                )
                == TEST_SERVICE_ID
            )

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id="overridden-project-id",
            paging_token=None,
        )


def test_terminate_service(setup_service_controller):
    """Test that terminate service"""
    service_controller, fake_sdk = setup_service_controller

    service_controller.terminate("service2_abc123")
    assert fake_sdk.terminated_service_id == "service2_abc123"


class TestListServices:
    def test_list_service_v2_with_id(self, setup_service_controller):
        service_controller, _ = setup_service_controller
        service_controller.list(service_id="service2_id")

        service_controller.api_client.get_service_api_v2_services_v2_service_id_get.assert_called_once_with(
            "service2_id",
        )

    @pytest.mark.parametrize("name", [None, "service_name"])
    @pytest.mark.parametrize("project_id", [None, "test_project_id"])
    @pytest.mark.parametrize("created_by_me", [True, False])
    def test_list_service(
        self,
        setup_service_controller,
        name: Optional[str],
        project_id: Optional[str],
        created_by_me: bool,
    ):
        service_controller, _ = setup_service_controller
        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[], metadata=Mock(next_paging_token=None),
        )
        if created_by_me:
            creator_id = "test_user_id"
            service_controller.api_client.get_user_info_api_v2_userinfo_get.return_value = Mock(
                result=Mock(id=creator_id)
            )
        else:
            creator_id = None

        service_controller.list(
            name=name, project_id=project_id, created_by_me=created_by_me
        )

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            creator_id=creator_id,
            name=name,
            project_id=project_id,
            count=10,
            paging_token=None,
        )

    def test_list_service_with_pagination(
        self, setup_service_controller,
    ):
        service_controller, _ = setup_service_controller

        next_paging_token = "test"
        list_return_values = [
            DecoratedlistserviceapimodelListResponse(
                results=[],
                metadata=ListResponseMetadata(
                    total=11, next_paging_token=next_paging_token
                ),
            ),
            DecoratedlistserviceapimodelListResponse(
                results=[],
                metadata=ListResponseMetadata(total=11, next_paging_token=None),
            ),
        ]
        service_controller.api_client.list_services_api_v2_services_v2_get.side_effect = (
            list_return_values
        )

        service_controller.list()

        calls = [
            call(
                creator_id=None, name=None, project_id=None, count=10, paging_token=None
            ),
            call(
                creator_id=None,
                name=None,
                project_id=None,
                count=10,
                paging_token=next_paging_token,
            ),
        ]
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_has_calls(
            calls
        )


def test_rollback(setup_service_controller):
    service_controller, fake_sdk = setup_service_controller

    assert fake_sdk.rolled_back_service_id is None
    assert fake_sdk.rolled_back_service_model is None

    # If max_surge_percent isn't provided, pass in RollbackServiceModel without it
    service_controller.rollback("service2_abc123")
    assert fake_sdk.rolled_back_service_id == "service2_abc123"
    assert fake_sdk.rolled_back_service_model.max_surge_percent is None
    fake_sdk.reset()

    # The behavior should be the same if it's explicitly None.
    service_controller.rollback("service2_abc123", None)
    assert fake_sdk.rolled_back_service_id == "service2_abc123"
    assert fake_sdk.rolled_back_service_model.max_surge_percent is None
    fake_sdk.reset()

    # If it's provided, pass it in through the RollbackServiceModel
    service_controller.rollback("service2_abc123", 43)
    assert fake_sdk.rolled_back_service_id == "service2_abc123"
    assert fake_sdk.rolled_back_service_model.max_surge_percent == 43
    fake_sdk.reset()


@pytest.mark.parametrize("init_runtime_env", [None, {}, {"working_dir": "."}])
def test_override_runtime_env_config_and_infer_project_id_v2(
    setup_service_controller, init_runtime_env: Collection[str],
):
    """Test helper functions in rollout function (service v2 path)
    Test that runtime_env is overwrriten from override_runtime_env_config
    Test the project_id is autopopulated from infer_project_id
    """
    config_dict = {
        "name": "mock_name",
        "description": "mock_description",
        "build_id": "mock_build_id",
        "compute_config_id": "mock_compute_config_id",
        "ray_serve_config": {
            "import_path": "mock_import_path",
            "runtime_env": init_runtime_env,
        },
    }

    assert "project_id" not in config_dict

    mock_project_id = "default"

    new_runtime_env = {"working_dir": "s3://bucket"}
    mock_override_config = Mock(return_value=new_runtime_env)
    mock_infer_project_id = Mock(return_value=mock_project_id)

    with patch.multiple(
        "anyscale.controllers.service_controller",
        infer_project_id=mock_infer_project_id,
        override_runtime_env_config=mock_override_config,
    ), patch.multiple(
        "anyscale.models.job_model", validate_successful_build=Mock(),
    ):
        service_controller, fake_sdk = setup_service_controller
        service_controller.api_client = Mock()
        service_controller.anyscale_api_client = Mock()
        service_controller.log = Mock()
        service_controller.api_client.apply_service_api_v2_services_v2_apply_put = (
            Mock()
        )

        service_controller.rollout(
            config_dict, auto_complete_rollout=True,
        )

        mock_override_config.assert_called_once_with(
            anyscale_api_client=fake_sdk,
            api_client=service_controller.api_client,
            compute_config_id=config_dict["compute_config_id"],
            runtime_env=init_runtime_env,
            workload_type=Workload.SERVICES,
            log=service_controller.log,
        )

        expected_ray_serve_config = {
            "import_path": "mock_import_path",
            "runtime_env": new_runtime_env,
        }

        assert fake_sdk.rolled_out_service_model == ApplyServiceModel(
            name=config_dict["name"],
            description=config_dict["description"],
            project_id=mock_project_id,
            canary_percent=None,
            ray_serve_config=expected_ray_serve_config,
            build_id=config_dict["build_id"],
            compute_config_id=config_dict["compute_config_id"],
            rollout_strategy=None,
            auto_complete_rollout=True,
            max_surge_percent=None,
        )
        fake_sdk.reset()


def test_overwrite_runtime_env_in_v2_ray_serve_config(setup_service_controller):
    """Test that runtime env is overwritten in v2 single-app configs as expected.
    Delegates the runtime env overwriting logic to override_runtime_env_config(), simply
    makes sure that override_runtime_env_config is called with the correct arguments.
    """
    service_controller, _ = setup_service_controller

    # Runtime environments
    original_runtime_env = {
        "pip": ["requests"],
        "working_dir": ".",
        "upload_path": "s3://already-a-uri/path",
    }
    rewritten_runtime_env = {
        "pip": ["requests"],
        "working_dir": "s3://bucket",
    }

    # Service configs
    with patch.multiple(
        "anyscale.models.job_model", validate_successful_build=Mock(),
    ):
        service_config = ServiceConfig.parse_obj(
            {
                "name": "mock_name",
                "description": "mock_description",
                "build_id": "mock_build_id",
                "compute_config_id": "mock_compute_config_id",
                "ray_serve_config": {
                    "import_path": "a.b.c",
                    "runtime_env": original_runtime_env,
                },
            }
        )
        expected_service_config = deepcopy(service_config)
        expected_service_config.ray_serve_config["runtime_env"] = rewritten_runtime_env

    # Perform test
    override_runtime_env_config_mock = Mock(return_value=rewritten_runtime_env)
    with patch.multiple(
        "anyscale.controllers.service_controller",
        override_runtime_env_config=override_runtime_env_config_mock,
    ):
        service_controller._overwrite_runtime_env_in_v2_ray_serve_config(service_config)

    calls = [
        call(
            runtime_env=original_runtime_env,
            anyscale_api_client=ANY,
            api_client=ANY,
            workload_type=Workload.SERVICES,
            compute_config_id="mock_compute_config_id",
            log=ANY,
        )
    ]
    assert service_config == expected_service_config
    override_runtime_env_config_mock.assert_has_calls(calls)


def test_overwrite_runtime_env_in_v2_ray_serve_config_multi_app(
    setup_service_controller,
):
    """Test that runtime env for each app is overwritten in v2 multi-app configs.
    Delegates the runtime env overwriting logic to override_runtime_env_config(), simply
    makes sure that override_runtime_env_config is called with the correct arguments.
    """
    service_controller, _ = setup_service_controller

    # Runtime environments
    original_runtime_env1 = {
        "pip": ["requests"],
        "working_dir": ".",
        "upload_path": "s3://already-a-uri/path",
    }
    original_runtime_env2 = None
    rewritten_runtime_env = {
        "pip": ["requests"],
        "working_dir": "s3://bucket",
    }

    # Service configs
    with patch.multiple(
        "anyscale.models.job_model", validate_successful_build=Mock(),
    ):
        service_config = ServiceConfig.parse_obj(
            {
                "name": "mock_name",
                "description": "mock_description",
                "build_id": "mock_build_id",
                "compute_config_id": "mock_compute_config_id",
                "ray_serve_config": {
                    "applications": [
                        {"import_path": "a.b", "runtime_env": original_runtime_env1},
                        {"import_path": "c.d"},
                    ]
                },
            }
        )
        expected_service_config = deepcopy(service_config)
        for app in expected_service_config.ray_serve_config["applications"]:
            app["runtime_env"] = rewritten_runtime_env

    # Perform test
    override_runtime_env_config_mock = Mock(return_value=rewritten_runtime_env)
    with patch.multiple(
        "anyscale.controllers.service_controller",
        override_runtime_env_config=override_runtime_env_config_mock,
    ):
        service_controller._overwrite_runtime_env_in_v2_ray_serve_config(service_config)

    calls = [
        call(
            runtime_env=original_runtime_env1,
            anyscale_api_client=ANY,
            api_client=ANY,
            workload_type=Workload.SERVICES,
            compute_config_id="mock_compute_config_id",
            log=ANY,
        ),
        call(
            runtime_env=original_runtime_env2,
            anyscale_api_client=ANY,
            api_client=ANY,
            workload_type=Workload.SERVICES,
            compute_config_id="mock_compute_config_id",
            log=ANY,
        ),
    ]
    assert service_config == expected_service_config
    override_runtime_env_config_mock.assert_has_calls(calls)



from typing import Any, Dict, Optional, Union

from anyscale._private.sdk import sdk_command
from anyscale.cli_logger import BlockLogger
from anyscale.job._private.job_sdk import PrivateJobSDK
from anyscale.job.models import JobConfig, JobLogMode, JobState, JobStatus


logger = BlockLogger()


def _resolve_id_from_args(
    id: Optional[str], kwargs: Dict[str, Any]  # noqa: A002
) -> Optional[str]:
    """Return the correct id as passed through id and kwargs.

    As job_id is being soft deprecated, we will warn if that is passed
    through kwargs.

    If id is passed, id will always be returned (regardless of job_id
    being passed in kwargs). If id is None and job_id is passed in kwargs,
    we will return that as the id to be used.
    """
    if "job_id" in kwargs:
        logger.warning("`job_id` has been deprecated, use `id` instead.")

    if id is not None:
        return id
    else:
        return kwargs.get("job_id", None)


_JOB_SDK_SINGLETON_KEY = "job_sdk"

_SUBMIT_EXAMPLE = """
import anyscale
from anyscale.job.models import JobConfig

anyscale.job.submit(
    JobConfig(
        name="my-job",
        entrypoint="python main.py",
        working_dir=".",
    ),
)
"""

_SUBMIT_ARG_DOCSTRINGS = {"config": "The config options defining the job."}


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    PrivateJobSDK,
    doc_py_example=_SUBMIT_EXAMPLE,
    arg_docstrings=_SUBMIT_ARG_DOCSTRINGS,
)
def submit(config: JobConfig, *, _sdk: PrivateJobSDK) -> str:
    """Submit a job.

    Returns the id of the submitted job.
    """
    return _sdk.submit(config)


_STATUS_EXAMPLE = """
import anyscale
from anyscale.job.models import JobStatus

status: JobStatus = anyscale.job.status(name="my-job")
"""

_STATUS_ARG_DOCSTRINGS = {
    "name": "Name of the job.",
    "id": "Unique ID of the job",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    PrivateJobSDK,
    doc_py_example=_STATUS_EXAMPLE,
    arg_docstrings=_STATUS_ARG_DOCSTRINGS,
)
def status(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _sdk: PrivateJobSDK,
    **_kwargs: Dict[str, Any],
) -> JobStatus:
    """Get the status of a job."""
    id = _resolve_id_from_args(id, _kwargs)  # noqa: A001
    return _sdk.status(name=name, job_id=id, cloud=cloud, project=project)


_TERMINATE_EXAMPLE = """
import anyscale

anyscale.job.terminate(name="my-job")
"""

_TERMINATE_ARG_DOCSTRINGS = {
    "name": "Name of the job.",
    "id": "Unique ID of the job",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    PrivateJobSDK,
    doc_py_example=_TERMINATE_EXAMPLE,
    arg_docstrings=_TERMINATE_ARG_DOCSTRINGS,
)
def terminate(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _sdk: PrivateJobSDK,
    **_kwargs: Dict[str, Any],
) -> str:
    """Terminate a job.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the terminated job.
    """
    id = _resolve_id_from_args(id, _kwargs)  # noqa: A001
    return _sdk.terminate(name=name, job_id=id, cloud=cloud, project=project)


_ARCHIVE_EXAMPLE = """
import anyscale

anyscale.job.archive(name="my-job")
"""

_ARCHIVE_ARG_DOCSTRINGS = {
    "name": "Name of the job.",
    "id": "Unique ID of the job",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the job . If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    PrivateJobSDK,
    doc_py_example=_ARCHIVE_EXAMPLE,
    arg_docstrings=_ARCHIVE_ARG_DOCSTRINGS,
)
def archive(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _sdk: PrivateJobSDK,
    **_kwargs: Dict[str, Any],
) -> str:
    """Archive a job.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the archived job.
    """
    id = _resolve_id_from_args(id, _kwargs)  # noqa: A001
    return _sdk.archive(name=name, job_id=id, cloud=cloud, project=project)


_WAIT_EXAMPLE = """\
import anyscale

anyscale.job.wait(name="my-job", timeout_s=180)"""

_WAIT_ARG_DOCSTRINGS = {
    "name": "Name of the job.",
    "id": "Unique ID of the job",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
    "state": "Target state of the job",
    "timeout_s": "Number of seconds to wait before timing out, this timeout will not affect job execution",
}


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    PrivateJobSDK,
    doc_py_example=_WAIT_EXAMPLE,
    arg_docstrings=_WAIT_ARG_DOCSTRINGS,
)
def wait(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    state: Union[JobState, str] = JobState.SUCCEEDED,
    timeout_s: float = 1800,
    _sdk: PrivateJobSDK,
    **_kwargs: Dict[str, Any],
):
    """"Wait for a job to enter a specific state."""
    id = _resolve_id_from_args(id, _kwargs)  # noqa: A001
    _sdk.wait(
        name=name,
        job_id=id,
        cloud=cloud,
        project=project,
        state=state,
        timeout_s=timeout_s,
    )


_GET_LOGS_EXAMPLE = """\
import anyscale

anyscale.job.get_logs(name="my-job", run="job-run-name")
"""

_GET_LOGS_ARG_DOCSTRINGS = {
    "name": "Name of the job",
    "id": "Unique ID of the job",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
    "run": "The name of the run to query. Names can be found in the JobStatus. If not provided, the last job run will be used.",
    "mode": "The mode of log fetching to be used. Supported modes can be found in JobLogMode. If not provided, JobLogMode.TAIL will be used.",
    "max_lines": "The number of log lines to be fetched. If not provided, the complete log will be fetched.",
}


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    PrivateJobSDK,
    doc_py_example=_GET_LOGS_EXAMPLE,
    arg_docstrings=_GET_LOGS_ARG_DOCSTRINGS,
)
def get_logs(
    *,
    id: Optional[str] = None,  # noqa: A002
    name: Optional[str] = None,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    run: Optional[str] = None,
    mode: Union[str, JobLogMode] = JobLogMode.TAIL,
    max_lines: Optional[int] = None,
    _sdk: PrivateJobSDK,
    **_kwargs: Dict[str, Any],
) -> str:
    """Query the jobs for a job run."""
    id = _resolve_id_from_args(id, _kwargs)  # noqa: A001
    return _sdk.get_logs(
        job_id=id,
        name=name,
        cloud=cloud,
        project=project,
        run=run,
        mode=mode,
        max_lines=max_lines,
    )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreateFineTuningJobProductRequest(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'training_file': 'str',
        'validation_file': 'str',
        'hyperparameters': 'CreateFineTuningHyperparameters',
        'model': 'str',
        'suffix': 'str',
        'cloud_id': 'str',
        'version': 'str',
        'instance_type': 'str',
        'batch_size_per_device': 'int'
    }

    attribute_map = {
        'training_file': 'training_file',
        'validation_file': 'validation_file',
        'hyperparameters': 'hyperparameters',
        'model': 'model',
        'suffix': 'suffix',
        'cloud_id': 'cloud_id',
        'version': 'version',
        'instance_type': 'instance_type',
        'batch_size_per_device': 'batch_size_per_device'
    }

    def __init__(self, training_file=None, validation_file=None, hyperparameters=None, model=None, suffix=None, cloud_id=None, version=None, instance_type=None, batch_size_per_device=None, local_vars_configuration=None):  # noqa: E501
        """CreateFineTuningJobProductRequest - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._training_file = None
        self._validation_file = None
        self._hyperparameters = None
        self._model = None
        self._suffix = None
        self._cloud_id = None
        self._version = None
        self._instance_type = None
        self._batch_size_per_device = None
        self.discriminator = None

        self.training_file = training_file
        if validation_file is not None:
            self.validation_file = validation_file
        if hyperparameters is not None:
            self.hyperparameters = hyperparameters
        self.model = model
        if suffix is not None:
            self.suffix = suffix
        self.cloud_id = cloud_id
        if version is not None:
            self.version = version
        if instance_type is not None:
            self.instance_type = instance_type
        if batch_size_per_device is not None:
            self.batch_size_per_device = batch_size_per_device

    @property
    def training_file(self):
        """Gets the training_file of this CreateFineTuningJobProductRequest.  # noqa: E501

        ID of the .jsonl file used for training  # noqa: E501

        :return: The training_file of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._training_file

    @training_file.setter
    def training_file(self, training_file):
        """Sets the training_file of this CreateFineTuningJobProductRequest.

        ID of the .jsonl file used for training  # noqa: E501

        :param training_file: The training_file of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and training_file is None:  # noqa: E501
            raise ValueError("Invalid value for `training_file`, must not be `None`")  # noqa: E501

        self._training_file = training_file

    @property
    def validation_file(self):
        """Gets the validation_file of this CreateFineTuningJobProductRequest.  # noqa: E501

        ID of the .jsonl file used for validation. Optional  # noqa: E501

        :return: The validation_file of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._validation_file

    @validation_file.setter
    def validation_file(self, validation_file):
        """Sets the validation_file of this CreateFineTuningJobProductRequest.

        ID of the .jsonl file used for validation. Optional  # noqa: E501

        :param validation_file: The validation_file of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """

        self._validation_file = validation_file

    @property
    def hyperparameters(self):
        """Gets the hyperparameters of this CreateFineTuningJobProductRequest.  # noqa: E501

        Hyperparameters used for fine-tuning.  # noqa: E501

        :return: The hyperparameters of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: CreateFineTuningHyperparameters
        """
        return self._hyperparameters

    @hyperparameters.setter
    def hyperparameters(self, hyperparameters):
        """Sets the hyperparameters of this CreateFineTuningJobProductRequest.

        Hyperparameters used for fine-tuning.  # noqa: E501

        :param hyperparameters: The hyperparameters of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: CreateFineTuningHyperparameters
        """

        self._hyperparameters = hyperparameters

    @property
    def model(self):
        """Gets the model of this CreateFineTuningJobProductRequest.  # noqa: E501

        A supported model used as a base for fine-tuning  # noqa: E501

        :return: The model of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._model

    @model.setter
    def model(self, model):
        """Sets the model of this CreateFineTuningJobProductRequest.

        A supported model used as a base for fine-tuning  # noqa: E501

        :param model: The model of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and model is None:  # noqa: E501
            raise ValueError("Invalid value for `model`, must not be `None`")  # noqa: E501

        self._model = model

    @property
    def suffix(self):
        """Gets the suffix of this CreateFineTuningJobProductRequest.  # noqa: E501

        A suffix (along with a hash) that will be appended to the model to create `fine_tuned_model`. Optional.  # noqa: E501

        :return: The suffix of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._suffix

    @suffix.setter
    def suffix(self, suffix):
        """Sets the suffix of this CreateFineTuningJobProductRequest.

        A suffix (along with a hash) that will be appended to the model to create `fine_tuned_model`. Optional.  # noqa: E501

        :param suffix: The suffix of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """

        self._suffix = suffix

    @property
    def cloud_id(self):
        """Gets the cloud_id of this CreateFineTuningJobProductRequest.  # noqa: E501


        :return: The cloud_id of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this CreateFineTuningJobProductRequest.


        :param cloud_id: The cloud_id of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def version(self):
        """Gets the version of this CreateFineTuningJobProductRequest.  # noqa: E501

        The version of llmforge to use. If not specified, the latest version will be used. This version dictates the minimum Endpoint version needed to serve the produced fine-tuned model.  # noqa: E501

        :return: The version of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this CreateFineTuningJobProductRequest.

        The version of llmforge to use. If not specified, the latest version will be used. This version dictates the minimum Endpoint version needed to serve the produced fine-tuned model.  # noqa: E501

        :param version: The version of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """

        self._version = version

    @property
    def instance_type(self):
        """Gets the instance_type of this CreateFineTuningJobProductRequest.  # noqa: E501

        The instance type to use for the fine-tuning job.  # noqa: E501

        :return: The instance_type of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._instance_type

    @instance_type.setter
    def instance_type(self, instance_type):
        """Sets the instance_type of this CreateFineTuningJobProductRequest.

        The instance type to use for the fine-tuning job.  # noqa: E501

        :param instance_type: The instance_type of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """
        allowed_values = ["p4de.24xlarge", "p4d.24xlarge", "g5.12xlarge", "g5.4xlarge"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and instance_type not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `instance_type` ({0}), must be one of {1}"  # noqa: E501
                .format(instance_type, allowed_values)
            )

        self._instance_type = instance_type

    @property
    def batch_size_per_device(self):
        """Gets the batch_size_per_device of this CreateFineTuningJobProductRequest.  # noqa: E501

        The number of samples that are processed by the model in a single iteration in training and evaluation.  # noqa: E501

        :return: The batch_size_per_device of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: int
        """
        return self._batch_size_per_device

    @batch_size_per_device.setter
    def batch_size_per_device(self, batch_size_per_device):
        """Sets the batch_size_per_device of this CreateFineTuningJobProductRequest.

        The number of samples that are processed by the model in a single iteration in training and evaluation.  # noqa: E501

        :param batch_size_per_device: The batch_size_per_device of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: int
        """

        self._batch_size_per_device = batch_size_per_device

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateFineTuningJobProductRequest):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateFineTuningJobProductRequest):
            return True

        return self.to_dict() != other.to_dict()



from dataclasses import dataclass, field
import re
from typing import Dict, Optional, Union

import pytest

from anyscale._private.models import ModelBase, ModelEnum


def test_invalid_subclass():
    """Validate that the subclass is a dataclass with frozen=True."""

    class NotADataclass(ModelBase):
        pass

    with pytest.raises(
        TypeError,
        match="Subclasses of `ModelBase` must be dataclasses with `frozen=True`.",
    ):
        NotADataclass()

    @dataclass
    class NotFrozen(ModelBase):
        pass

    with pytest.raises(
        TypeError,
        match="Subclasses of `ModelBase` must be dataclasses with `frozen=True`.",
    ):
        NotFrozen()


class TestValidatorValidation:
    def test_missing(self):
        @dataclass(frozen=True)
        class Subclass(ModelBase):
            some_field: str = field(metadata={"docstring": "test"})
            some_other_field: str = field(metadata={"docstring": "test"})

            def _validate_some_field(self, some_field: str):
                pass

        with pytest.raises(
            RuntimeError,
            match="Model 'Subclass' is missing validator method for field 'some_other_field'.",
        ):
            Subclass(some_field="hi", some_other_field="hi2")

    def test_multiple_params(self):
        @dataclass(frozen=True)
        class Subclass(ModelBase):
            some_field: str = field(metadata={"docstring": "test"})

            def _validate_some_field(self, some_field: str, something_else: int):
                pass

        with pytest.raises(
            TypeError,
            match=re.escape(
                "Validator for field 'some_field' must take the field value as its only argument."
            ),
        ):
            Subclass(some_field="hi")

    def test_wrong_param_name(self):
        @dataclass(frozen=True)
        class Subclass(ModelBase):
            some_field: str = field(metadata={"docstring": "test"})

            def _validate_some_field(self, arg: str):
                pass

        with pytest.raises(
            TypeError,
            match=re.escape(
                "Validator for field 'some_field' has invalid parameter name (must match the field name)."
            ),
        ):
            Subclass(some_field="hi")

    def test_missing_param_annotation(self):
        @dataclass(frozen=True)
        class Subclass(ModelBase):
            some_field: str = field(metadata={"docstring": "test"})

            def _validate_some_field(self, some_field):
                pass

        with pytest.raises(
            TypeError,
            match=re.escape(
                "Validator for field 'some_field' is missing a type annotation"
            ),
        ):
            Subclass(some_field="hi")

    def test_wrong_param_annotation(self):
        @dataclass(frozen=True)
        class Subclass(ModelBase):
            some_field: str = field(metadata={"docstring": "test"})

            def _validate_some_field(self, some_field: int):
                pass

        with pytest.raises(
            TypeError,
            match=re.escape(
                "Validator for field 'some_field' has the wrong type annotation"
            ),
        ):
            Subclass(some_field="hi")


def test_validators():
    @dataclass(frozen=True)
    class Subclass(ModelBase):
        some_field: str = field(metadata={"docstring": "test"})
        some_other_field: str = field(metadata={"docstring": "test"})

        def _validate_some_field(self, some_field: str):
            if not isinstance(some_field, str):
                raise TypeError("'some_field' must be a string.")

        def _validate_some_other_field(self, some_other_field: str):
            if not isinstance(some_other_field, str):
                raise TypeError("'some_other_field' must be a string.")

    Subclass(some_field="hi", some_other_field="hi2")
    with pytest.raises(TypeError, match="'some_other_field' must be a string."):
        Subclass(some_field="hi", some_other_field=123)


def test_update_value_from_validator():
    @dataclass(frozen=True)
    class Subclass(ModelBase):
        some_field: str = field(metadata={"docstring": "test"})

        def _validate_some_field(self, some_field: str) -> str:
            if not isinstance(some_field, str):
                raise TypeError("'some_field' must be a string.")

            return some_field + "-updated"

    assert Subclass(some_field="hi").some_field == "hi-updated"


class TestToDict:
    def test_exclude_none(self):
        @dataclass(frozen=True)
        class SubModel(ModelBase):
            sub_field_1: Optional[str] = field(
                metadata={"detailed": True, "docstring": "test"}
            )
            sub_field_2: Optional[str] = field(metadata={"docstring": "test"})

            def _validate_sub_field_1(self, sub_field_1: Optional[str]):
                pass

            def _validate_sub_field_2(self, sub_field_2: Optional[str]):
                pass

        @dataclass(frozen=True)
        class Model(ModelBase):
            sub_model: Optional[SubModel] = field(metadata={"docstring": "test"})

            def _validate_sub_model(self, sub_model: Optional[SubModel]):
                pass

        assert Model(sub_model=None).to_dict() == {}
        assert Model(sub_model=None).to_dict(exclude_none=False) == {"sub_model": None}
        assert Model(
            sub_model=SubModel(sub_field_1=None, sub_field_2="hi")
        ).to_dict() == {"sub_model": {"sub_field_2": "hi"},}
        assert Model(sub_model=SubModel(sub_field_1=None, sub_field_2="hi")).to_dict(
            exclude_none=False
        ) == {"sub_model": {"sub_field_1": None, "sub_field_2": "hi"},}


class TestFromDict:
    def test_basic(self):
        @dataclass(frozen=True)
        class Model(ModelBase):
            field_1: str = field(metadata={"docstring": "test"})
            field_2: Optional[int] = field(default=None, metadata={"docstring": "test"})

            def _validate_field_1(self, field_1: str):
                pass

            def _validate_field_2(self, field_2: Optional[int]):
                pass

        assert Model.from_dict({"field_1": "hi"},) == Model(field_1="hi")

        assert Model.from_dict({"field_1": "hi", "field_2": 1},) == Model(
            field_1="hi", field_2=1
        )

        with pytest.raises(
            TypeError, match="missing 1 required positional argument: 'field_1'"
        ):
            Model.from_dict({"field_2": 1},)

    def test_nested(self):
        @dataclass(frozen=True)
        class SubModel(ModelBase):
            sub_field: int = field(metadata={"docstring": "test"})

            def _validate_sub_field(self, sub_field: int):
                pass

        @dataclass(frozen=True)
        class Model(ModelBase):
            field_1: str = field(metadata={"docstring": "test"})
            sub_model: Union[Dict, SubModel] = field(metadata={"docstring": "test"})
            optional_sub_model: Union[None, Dict, SubModel] = field(
                default=None, metadata={"docstring": "test"}
            )

            def _validate_field_1(self, field_1: str):
                pass

            def _validate_sub_model(self, sub_model: Union[Dict, SubModel]) -> SubModel:
                if isinstance(sub_model, dict):
                    sub_model = SubModel.from_dict(sub_model)

                if not isinstance(sub_model, SubModel):
                    raise TypeError("'sub_model' must be SubModel or dict")

                return sub_model

            def _validate_optional_sub_model(
                self, optional_sub_model: Union[None, Dict, SubModel]
            ) -> Optional[SubModel]:
                if optional_sub_model is None:
                    return optional_sub_model

                if isinstance(optional_sub_model, dict):
                    optional_sub_model = SubModel.from_dict(optional_sub_model)

                if not isinstance(optional_sub_model, SubModel):
                    raise TypeError("'optional_sub_model' must be SubModel or dict")

                return optional_sub_model

        assert Model.from_dict(
            {"field_1": "hi", "sub_model": {"sub_field": 1}},
        ) == Model(field_1="hi", sub_model=SubModel(sub_field=1))

        assert Model.from_dict(
            {"field_1": "hi", "sub_model": SubModel(sub_field=1)},
        ) == Model(field_1="hi", sub_model=SubModel(sub_field=1))

        assert Model.from_dict(
            {
                "field_1": "hi",
                "sub_model": SubModel(sub_field=1),
                "optional_sub_model": SubModel(sub_field=1),
            },
        ) == Model(
            field_1="hi",
            sub_model=SubModel(sub_field=1),
            optional_sub_model=SubModel(sub_field=1),
        )

        assert Model.from_dict(
            {
                "field_1": "hi",
                "sub_model": SubModel(sub_field=1),
                "optional_sub_model": {"sub_field": 1},
            },
        ) == Model(
            field_1="hi",
            sub_model=SubModel(sub_field=1),
            optional_sub_model=SubModel(sub_field=1),
        )


def test_options():
    @dataclass(frozen=True)
    class Model(ModelBase):
        field_1: Optional[str] = field(metadata={"docstring": "test"})
        field_2: Optional[str] = field(metadata={"docstring": "test"})

        def _validate_field_1(self, field_1: Optional[str]):
            pass

        def _validate_field_2(self, field_2: Optional[str]):
            pass

    m0 = Model(field_1=None, field_2="hi2")
    assert m0.field_1 is None
    assert m0.field_2 == "hi2"

    # Check that only the passed value is overwritten.
    # m0 should be unmodified.
    m1 = m0.options(field_1="hi1")
    assert m1.field_1 == "hi1"
    assert m1.field_2 == "hi2"
    assert m0.field_1 is None
    assert m0.field_2 == "hi2"

    # Check that `None` can be passed to overwrite a value.
    m2 = m0.options(field_2=None)
    assert m2.field_1 is None
    assert m2.field_2 is None
    assert m0.field_1 is None
    assert m0.field_2 == "hi2"

    # Test passing an unknown field.
    with pytest.raises(
        ValueError,
        match=re.escape("Unexpected values passed to '.options': ['unknown']."),
    ):
        m0.options(unknown="oops")


class TestModelEnum:
    def test_missing_docstrings_value(self):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "ModelEnum 'CustomEnum.__docstrings__' is missing docstrings for values: ['VAL']"
            ),
        ):

            class CustomEnum(ModelEnum):
                VAL = "VAL"

    def test_bad_docstrings_type(self):
        with pytest.raises(
            TypeError,
            match=re.escape("ModelEnum 'CustomEnum.__docstrings__' is the wrong type"),
        ):

            class CustomEnum(ModelEnum):
                VAL = "VAL"

                __docstrings__ = {VAL: 123}

    def test_convert_to_str(self):
        class CustomEnum(ModelEnum):
            VAL = "VAL"

            __docstrings__ = {VAL: "test"}

        @dataclass(frozen=True)
        class Model(ModelBase):
            e: CustomEnum = field(metadata={"docstring": "test"})

            def _validate_e(self, e: CustomEnum):
                pass

        assert Model(e=CustomEnum.VAL).to_dict() == {"e": "VAL"}



from dataclasses import asdict, fields
from enum import Enum, EnumMeta
import inspect
from typing import Any, Dict, Iterable, Tuple, TypeVar, Union

import yaml


# In Python 3.11 there is a 'Self' introduced, but this is the
# workaround before that: https://peps.python.org/pep-0673/.
TModelBase = TypeVar("TModelBase", bound="ModelBase")


class ModelEnumType(EnumMeta):
    def __new__(cls, subcls, bases, classdict):
        cls = super().__new__(cls, subcls, bases, classdict)

        # Assert that all enum values have docstrings.
        if not isinstance(cls.__docstrings__, dict) or not all(
            isinstance(k, str) and isinstance(v, str)
            for k, v in cls.__docstrings__.items()
        ):
            raise TypeError(
                f"ModelEnum '{cls.__name__}.__docstrings__' is the wrong type. "
                "Must be a Dict[ModelEnum, str] documenting all enum values"
            )

        missing = {v.value for v in cls if not v.name.startswith("_")} - set(
            cls.__docstrings__.keys()
        )
        if missing:
            raise ValueError(
                f"ModelEnum '{cls.__name__}.__docstrings__' is missing docstrings for values: {sorted(missing)}"
            )

        return cls


class ModelEnum(str, Enum, metaclass=ModelEnumType):
    # Must be populated by subclasses to document every value.
    __docstrings__: Dict["ModelEnum", str] = {}

    def __str__(self):
        return self.name

    @classmethod
    def validate(cls, value: Union[str, "ModelEnum"]) -> "ModelEnum":
        allowed_values = [str(s) for s in cls.__members__ if not str(s).startswith("_")]
        if value.upper() in allowed_values:
            return cls(value.upper())
        else:
            raise ValueError(
                f"'{value.upper()}' is not a valid {cls.__name__}. Allowed values are {allowed_values}"
            )


class ModelBaseType(type):
    pass


class ModelBase(metaclass=ModelBaseType):
    def __new__(cls, *args, **kwargs):  # noqa: ARG003
        """Validate that the subclass is a conforming dataclass."""
        if (
            not hasattr(cls, "__dataclass_params__")
            or not cls.__dataclass_params__.frozen
        ):
            raise TypeError(
                "Subclasses of `ModelBase` must be dataclasses with `frozen=True`."
            )

        return super().__new__(cls)

    def _run_validators(self):
        """Run magic validator methods.

        Looks for methods with the name: `_validate_{field}`. The method will be called
        with the field value as its sole argument.

        If the validator has a return type annotation, the field will be updated to the
        returned value.
        """
        for field in fields(self):
            if not field.metadata.get("docstring", None):
                raise ValueError(
                    f"Model '{type(self).__name__}' is missing docstring for field '{field.name}'."
                )

            validator = getattr(self, f"_validate_{field.name}", None)
            if validator is None:
                raise RuntimeError(
                    f"Model '{type(self).__name__}' is missing validator method for field '{field.name}'."
                )

            signature: inspect.Signature = inspect.signature(validator)
            if len(signature.parameters) != 1:
                raise TypeError(
                    f"Validator for field '{field.name}' must take the field value as its only argument."
                )
            elif field.name not in signature.parameters:
                raise TypeError(
                    f"Validator for field '{field.name}' has invalid parameter name (must match the field name)."
                )
            elif signature.parameters[field.name].annotation == inspect.Parameter.empty:
                raise TypeError(
                    f"Validator for field '{field.name}' is missing a type annotation (should be '{field.type}')."
                )
            elif signature.parameters[field.name].annotation != field.type:
                raise TypeError(
                    f"Validator for field '{field.name}' has the wrong type annotation (should be '{field.type}')."
                )

            maybe_updated_field_value = validator(getattr(self, field.name))
            if signature.return_annotation != inspect.Signature.empty:
                # Only update the value if the validator has a return type annotation.
                object.__setattr__(self, field.name, maybe_updated_field_value)

    def __post_init__(self):
        self._run_validators()

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "ModelBase":
        return cls(**d)

    @classmethod
    def from_yaml(cls, path: str, **kwargs) -> "ModelBase":
        with open(path) as f:
            args = yaml.safe_load(f) or {}
            if kwargs:
                args.update(kwargs)
            return cls.from_dict(args)

    def to_dict(self, *, exclude_none: bool = True) -> Dict[str, Any]:
        """Convert the model to a dictionary representation.

        If `exclude_none` is `True`, keys whose values are `None` will be excluded.
        """

        def maybe_exclude_nones(i: Iterable[Tuple[str, Any]]):
            d = {}
            for k, v in i:
                # Exclude none values if specified.
                if exclude_none and v is None:
                    continue

                # Convert enums to their string representation.
                if isinstance(v, ModelEnum):
                    v = v.value

                d[k] = v

            return d

        return asdict(self, dict_factory=maybe_exclude_nones)

    def options(self: TModelBase, **kwargs) -> TModelBase:
        """Return a copy of the model with the provided fields updated.

        All fields in the constructor are supported. Those not provided will be unchanged.
        """
        new_instance_kwargs = {
            field.name: kwargs.pop(field.name)
            if field.name in kwargs
            else getattr(self, field.name)
            for field in fields(self)
        }
        if len(kwargs) > 0:
            raise ValueError(
                f"Unexpected values passed to '.options': {list(kwargs.keys())}."
            )

        return type(self)(**new_instance_kwargs)



from functools import wraps
from typing import Callable, cast, Dict, Type, TypeVar

from typing_extensions import ParamSpec


P = ParamSpec("P")
T = TypeVar("T")


_LAZY_SDK_SINGLETONS: Dict[str, Callable] = {}


def sdk_command(
    key: str, sdk_cls: Type, *, doc_py_example: str, arg_docstrings: Dict[str, str],
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator to automatically inject an `_sdk` arg into the wrapped function.

    The arguments to this class are a unique key for the singleton and its type
    (the constructor will be called with no arguments).
    """

    # The P and T type hints allow f's type hints to pass through this decorator.
    # Without them, f's type hints would not be visible to the developer.
    # See https://github.com/anyscale/product/pull/27738.
    def _inject_typed_sdk_singleton(f: Callable[P, T]) -> Callable[P, T]:
        if not doc_py_example:
            raise ValueError(
                f"SDK command '{f.__name__}' must provide a non-empty 'doc_py_example'."
            )

        # TODO: validate docstrings.

        @wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            # We disable the mypy linter here because it treats kwargs as a
            # P.kwargs object. mypy wrongly thinks kwargs can't be indexed.
            if "_sdk" not in kwargs:  # type: ignore
                if key not in _LAZY_SDK_SINGLETONS:
                    _LAZY_SDK_SINGLETONS[key] = sdk_cls()

                kwargs["_sdk"] = _LAZY_SDK_SINGLETONS[key]  # type: ignore

            return f(*args, **kwargs)

        # TODO(edoakes): move to parsing docstrings instead.
        wrapper.__doc_py_example__ = doc_py_example  # type: ignore
        wrapper.__arg_docstrings__ = arg_docstrings  # type: ignore

        return wrapper

    return _inject_typed_sdk_singleton


def sdk_docs(
    *, doc_py_example: str, arg_docstrings: Dict[str, str],
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator to add documentation for an SDK command."""

    # The P and T type hints allow f's type hints to pass through this decorator.
    # Without them, f's type hints would not be visible to the developer.
    # See https://github.com/anyscale/product/pull/27738.
    def _add_doc_magic_attrs(f: Callable[P, T]) -> Callable[P, T]:
        if not doc_py_example:
            raise ValueError(
                f"SDK command '{f.__name__}' must provide a non-empty 'doc_py_example'."
            )

        # TODO(edoakes): validate docstrings.
        # TODO(edoakes): move to parsing docstrings instead.
        f.__doc_py_example__ = doc_py_example  # type: ignore
        f.__arg_docstrings__ = arg_docstrings  # type: ignore

        return f

    return _add_doc_magic_attrs



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedUnifiedJob(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'name': 'str',
        'job_type': 'UnifiedJobType',
        'created_at': 'datetime',
        'finished_at': 'datetime',
        'archived_at': 'datetime',
        'status': 'UnifiedJobStatus',
        'cost': 'float',
        'position_in_job_queue': 'int',
        'last_job_run_id': 'str',
        'project': 'MiniProject',
        'cluster': 'MiniCluster',
        'creator': 'MiniUser'
    }

    attribute_map = {
        'id': 'id',
        'name': 'name',
        'job_type': 'job_type',
        'created_at': 'created_at',
        'finished_at': 'finished_at',
        'archived_at': 'archived_at',
        'status': 'status',
        'cost': 'cost',
        'position_in_job_queue': 'position_in_job_queue',
        'last_job_run_id': 'last_job_run_id',
        'project': 'project',
        'cluster': 'cluster',
        'creator': 'creator'
    }

    def __init__(self, id=None, name=None, job_type=None, created_at=None, finished_at=None, archived_at=None, status=None, cost=None, position_in_job_queue=None, last_job_run_id=None, project=None, cluster=None, creator=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedUnifiedJob - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._name = None
        self._job_type = None
        self._created_at = None
        self._finished_at = None
        self._archived_at = None
        self._status = None
        self._cost = None
        self._position_in_job_queue = None
        self._last_job_run_id = None
        self._project = None
        self._cluster = None
        self._creator = None
        self.discriminator = None

        self.id = id
        self.name = name
        self.job_type = job_type
        self.created_at = created_at
        if finished_at is not None:
            self.finished_at = finished_at
        if archived_at is not None:
            self.archived_at = archived_at
        self.status = status
        self.cost = cost
        if position_in_job_queue is not None:
            self.position_in_job_queue = position_in_job_queue
        if last_job_run_id is not None:
            self.last_job_run_id = last_job_run_id
        self.project = project
        if cluster is not None:
            self.cluster = cluster
        self.creator = creator

    @property
    def id(self):
        """Gets the id of this DecoratedUnifiedJob.  # noqa: E501

        Server assigned unique identifier of the Job.  # noqa: E501

        :return: The id of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedUnifiedJob.

        Server assigned unique identifier of the Job.  # noqa: E501

        :param id: The id of this DecoratedUnifiedJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this DecoratedUnifiedJob.  # noqa: E501

        Name of the Job.  # noqa: E501

        :return: The name of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this DecoratedUnifiedJob.

        Name of the Job.  # noqa: E501

        :param name: The name of this DecoratedUnifiedJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def job_type(self):
        """Gets the job_type of this DecoratedUnifiedJob.  # noqa: E501

        The type of the Job.  # noqa: E501

        :return: The job_type of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: UnifiedJobType
        """
        return self._job_type

    @job_type.setter
    def job_type(self, job_type):
        """Sets the job_type of this DecoratedUnifiedJob.

        The type of the Job.  # noqa: E501

        :param job_type: The job_type of this DecoratedUnifiedJob.  # noqa: E501
        :type: UnifiedJobType
        """
        if self.local_vars_configuration.client_side_validation and job_type is None:  # noqa: E501
            raise ValueError("Invalid value for `job_type`, must not be `None`")  # noqa: E501

        self._job_type = job_type

    @property
    def created_at(self):
        """Gets the created_at of this DecoratedUnifiedJob.  # noqa: E501

        Time at which Job was created.  # noqa: E501

        :return: The created_at of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DecoratedUnifiedJob.

        Time at which Job was created.  # noqa: E501

        :param created_at: The created_at of this DecoratedUnifiedJob.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def finished_at(self):
        """Gets the finished_at of this DecoratedUnifiedJob.  # noqa: E501

        Time at which Job ended (completed or errored). If None, the job is still running.  # noqa: E501

        :return: The finished_at of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: datetime
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """Sets the finished_at of this DecoratedUnifiedJob.

        Time at which Job ended (completed or errored). If None, the job is still running.  # noqa: E501

        :param finished_at: The finished_at of this DecoratedUnifiedJob.  # noqa: E501
        :type: datetime
        """

        self._finished_at = finished_at

    @property
    def archived_at(self):
        """Gets the archived_at of this DecoratedUnifiedJob.  # noqa: E501

        Time the job was archived.  # noqa: E501

        :return: The archived_at of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: datetime
        """
        return self._archived_at

    @archived_at.setter
    def archived_at(self, archived_at):
        """Sets the archived_at of this DecoratedUnifiedJob.

        Time the job was archived.  # noqa: E501

        :param archived_at: The archived_at of this DecoratedUnifiedJob.  # noqa: E501
        :type: datetime
        """

        self._archived_at = archived_at

    @property
    def status(self):
        """Gets the status of this DecoratedUnifiedJob.  # noqa: E501

        The status of the Job.  # noqa: E501

        :return: The status of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: UnifiedJobStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this DecoratedUnifiedJob.

        The status of the Job.  # noqa: E501

        :param status: The status of this DecoratedUnifiedJob.  # noqa: E501
        :type: UnifiedJobStatus
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def cost(self):
        """Gets the cost of this DecoratedUnifiedJob.  # noqa: E501

        The cost of the Job.  # noqa: E501

        :return: The cost of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: float
        """
        return self._cost

    @cost.setter
    def cost(self, cost):
        """Sets the cost of this DecoratedUnifiedJob.

        The cost of the Job.  # noqa: E501

        :param cost: The cost of this DecoratedUnifiedJob.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and cost is None:  # noqa: E501
            raise ValueError("Invalid value for `cost`, must not be `None`")  # noqa: E501

        self._cost = cost

    @property
    def position_in_job_queue(self):
        """Gets the position_in_job_queue of this DecoratedUnifiedJob.  # noqa: E501

        The position of the job in the job queue. None for jobs not pending in a job queue.  # noqa: E501

        :return: The position_in_job_queue of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: int
        """
        return self._position_in_job_queue

    @position_in_job_queue.setter
    def position_in_job_queue(self, position_in_job_queue):
        """Sets the position_in_job_queue of this DecoratedUnifiedJob.

        The position of the job in the job queue. None for jobs not pending in a job queue.  # noqa: E501

        :param position_in_job_queue: The position_in_job_queue of this DecoratedUnifiedJob.  # noqa: E501
        :type: int
        """

        self._position_in_job_queue = position_in_job_queue

    @property
    def last_job_run_id(self):
        """Gets the last_job_run_id of this DecoratedUnifiedJob.  # noqa: E501

        The last job run ID of the job. None for interactive sessions.  # noqa: E501

        :return: The last_job_run_id of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: str
        """
        return self._last_job_run_id

    @last_job_run_id.setter
    def last_job_run_id(self, last_job_run_id):
        """Sets the last_job_run_id of this DecoratedUnifiedJob.

        The last job run ID of the job. None for interactive sessions.  # noqa: E501

        :param last_job_run_id: The last_job_run_id of this DecoratedUnifiedJob.  # noqa: E501
        :type: str
        """

        self._last_job_run_id = last_job_run_id

    @property
    def project(self):
        """Gets the project of this DecoratedUnifiedJob.  # noqa: E501

        The project in which this job lives  # noqa: E501

        :return: The project of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: MiniProject
        """
        return self._project

    @project.setter
    def project(self, project):
        """Sets the project of this DecoratedUnifiedJob.

        The project in which this job lives  # noqa: E501

        :param project: The project of this DecoratedUnifiedJob.  # noqa: E501
        :type: MiniProject
        """
        if self.local_vars_configuration.client_side_validation and project is None:  # noqa: E501
            raise ValueError("Invalid value for `project`, must not be `None`")  # noqa: E501

        self._project = project

    @property
    def cluster(self):
        """Gets the cluster of this DecoratedUnifiedJob.  # noqa: E501

        The cluster this job is running on  # noqa: E501

        :return: The cluster of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: MiniCluster
        """
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        """Sets the cluster of this DecoratedUnifiedJob.

        The cluster this job is running on  # noqa: E501

        :param cluster: The cluster of this DecoratedUnifiedJob.  # noqa: E501
        :type: MiniCluster
        """

        self._cluster = cluster

    @property
    def creator(self):
        """Gets the creator of this DecoratedUnifiedJob.  # noqa: E501

        The creator of this job  # noqa: E501

        :return: The creator of this DecoratedUnifiedJob.  # noqa: E501
        :rtype: MiniUser
        """
        return self._creator

    @creator.setter
    def creator(self, creator):
        """Sets the creator of this DecoratedUnifiedJob.

        The creator of this job  # noqa: E501

        :param creator: The creator of this DecoratedUnifiedJob.  # noqa: E501
        :type: MiniUser
        """
        if self.local_vars_configuration.client_side_validation and creator is None:  # noqa: E501
            raise ValueError("Invalid value for `creator`, must not be `None`")  # noqa: E501

        self._creator = creator

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedUnifiedJob):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedUnifiedJob):
            return True

        return self.to_dict() != other.to_dict()



from dataclasses import dataclass
import socket
import ssl
from typing import Callable, Dict, Iterator, Optional

import requests

from anyscale.cli_logger import BlockLogger


@dataclass
class DebugTestResult:
    name: str
    success: bool
    error: Optional[Exception] = None
    warnings: Optional[str] = None


def debug_cluster(
    logger: BlockLogger, cluster_domain_name: str, anyscale_ip: str
) -> Iterator[DebugTestResult]:
    has_experienced_a_failure: bool = False

    tests = list(_tests_for(cluster_domain_name, anyscale_ip).items())

    for name, test in tests:
        try:
            warning = test()

            yield DebugTestResult(name, True, warnings=warning)
        except Exception as e:  # noqa: BLE001
            yield DebugTestResult(name, False, e)

            if not has_experienced_a_failure:
                has_experienced_a_failure = True

                try:
                    cluster_ip_addr = socket.gethostbyname(cluster_domain_name)
                except Exception as e:  # noqa: BLE001
                    resp = requests.get(
                        "https://dns.google/resolve",
                        params={"name": cluster_domain_name},
                    )
                    resp.raise_for_status()

                    resp_json = resp.json()
                    cluster_ip_addr = resp_json["Answer"][0]["data"]
                logger.warning(
                    "Failure detected, retrying tests by directly communicating with the IP address."
                )
                tests.extend(
                    _tests_for(
                        cluster_ip_addr, anyscale_ip, tls_sni=cluster_domain_name
                    ).items()
                )


def _tests_for(
    endpoint: str, anyscale_ip: str, tls_sni: Optional[str] = None
) -> Dict[str, Callable[[], Optional[str]]]:
    """Attempt to debug networking for an Anyscale Cluster.

    Arguments:
        endpoint: Hostname or IP address of the cluster
        tls_sni: Hostname to use for TLS SNI (used if used to directly dial the IP address)

    Also ask for the IP from the control Plane

    Fallbacks:
    1. Try directly using different IPs + hostname

    """
    return {
        **(
            {
                "dns-lookup": lambda: _dns_test(endpoint, anyscale_ip),
                "dns-over-https": lambda: _dns_over_https_test(endpoint, anyscale_ip),
            }
            if tls_sni is None
            else {}
        ),
        "tcp-connect": lambda: _tcp_test(endpoint),
        "tls-no-sni": lambda: _tls_no_sni(endpoint),
        "tls-connect": lambda: _tls_test(endpoint, tls_sni),
        "https-get": lambda: _https_test(endpoint, tls_sni),
    }


def _dns_test(endpoint: str, anyscale_ip: str) -> Optional[str]:
    resolved_ip = socket.gethostbyname(endpoint)
    if resolved_ip != anyscale_ip:
        return (
            f"Unexpected local DNS resolution:\nLocal DNS resolves this to {resolved_ip}\n"
            f"Anyscale expects this host to resolve to {anyscale_ip}\n"
            "This is safe to ignore if you are using a VPN to connect to your cluster.\n"
        )
    return None


def _dns_over_https_test(endpoint: str, anyscale_ip: str) -> Optional[str]:
    resp = requests.get("https://dns.google/resolve", params={"name": endpoint})
    resp.raise_for_status()

    resp_json = resp.json()
    answers = resp_json.get("Answer", [])
    assert len(answers) > 0, "Got no answers for a successful query"

    resolved_ips = {answer["data"] for answer in answers}
    if anyscale_ip not in resolved_ips:
        return (
            f"Unexpected Public DNS resolution. This should never happen\n"
            f"`8.8.8.8` resolves this host to {resolved_ips}\n"
            f"Anyscale expects this host to resolve to {anyscale_ip}\n"
        )
    return None


def _tcp_test(endpoint: str) -> None:
    with socket.create_connection((endpoint, 443)):
        pass


def _tls_test(endpoint: str, hostname: Optional[str] = None) -> None:
    """Create a TLS connection to the destination.

        The optional hostname is used if the TLS connection is made directly to an IP address
        and we want to set the SNI separately.
    """
    if hostname is None:
        hostname = endpoint

    with socket.create_connection(
        (endpoint, 443)
    ) as sock, ssl.create_default_context().wrap_socket(sock, server_hostname=hostname):
        pass


def _tls_no_sni(endpoint: str) -> None:
    """Create a TLS connection to the destination without specifying an SNI.

    This is used to test if middleboxes/firewalls are rejecting connections based on SNI.
    """
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    with socket.create_connection((endpoint, 443)) as sock, ctx.wrap_socket(
        sock, server_hostname=None
    ):
        pass


def _https_test(endpoint: str, hostname: Optional[str] = None) -> None:
    if hostname is None:
        hostname = endpoint

    session = requests.Session()
    session.mount(f"https://{endpoint}", HostnameSpecificAdapter(hostname))
    resp = session.get(f"https://{endpoint}")

    if resp.status_code != requests.codes["unauthorized"]:
        raise RuntimeError(
            f"Unexpected status code from https://{endpoint}: {resp.status_code}"
        )

    assert (
        "x-anyscale-authentication-status" in resp.headers
        and resp.headers["x-anyscale-authentication-status"] == "Unauthorized"
    ), "Failed to get an expected anyscale-unique header from the response."


class HostnameSpecificAdapter(requests.adapters.HTTPAdapter):
    """Adapter to handle directly connecting to an IP address with a specific TLS Hostname.

    This allows requests to be made to https://<ip address>, and still have TLS work for the server name.
    """

    def __init__(self, hostname: str, *args, **kwargs):
        super().__init__(*args, **kwargs)

        class SingleHostSSLContext(ssl.SSLContext):
            def wrap_socket(self, *args, **kwargs):
                kwargs["server_hostname"] = hostname
                return super().wrap_socket(*args, **kwargs)

        self.init_poolmanager(
            self._pool_connections,  # type:ignore
            self._pool_maxsize,  # type:ignore
            ssl_context=SingleHostSSLContext(),
            assert_hostname=hostname,
        )



from dataclasses import dataclass
import os
import re
from typing import Optional

import pytest

from anyscale.compute_config.models import (
    ComputeConfig,
    ComputeConfigVersion,
    HeadNodeConfig,
    MarketType,
    WorkerNodeGroupConfig,
)


@dataclass
class ComputeConfigFile:
    name: str
    expected_config: Optional[ComputeConfig] = None
    expected_error: Optional[str] = None

    def get_path(self) -> str:
        return os.path.join(
            os.path.dirname(__file__), "test_files/compute_config_files", self.name
        )


TEST_CONFIG_FILES = [
    ComputeConfigFile(
        "minimal.yaml", expected_config=ComputeConfig(cloud="test-cloud-from-file"),
    ),
    ComputeConfigFile(
        "full.yaml",
        expected_config=ComputeConfig(
            cloud="test-cloud-from-file",
            zones=["zone1", "zone2"],
            enable_cross_zone_scaling=True,
            max_resources={"CPU": 10, "GPU": 1},
            head_node=HeadNodeConfig(
                instance_type="head-node-instance-type",
                resources={"CPU": 1},
                advanced_instance_config={"head": "config"},
            ),
            worker_nodes=[
                WorkerNodeGroupConfig(
                    name="worker-group-1",
                    instance_type="worker-group-1-instance-type",
                    resources={"CPU": 2},
                    min_nodes=2,
                    max_nodes=4,
                    market_type=MarketType.SPOT,
                    advanced_instance_config={"worker1": "config"},
                ),
                WorkerNodeGroupConfig(
                    name="worker-group-2",
                    instance_type="worker-group-2-instance-type",
                    resources={"GPU": 1, "CPU": 1},
                    min_nodes=0,
                    max_nodes=1,
                    market_type=MarketType.PREFER_SPOT,
                    advanced_instance_config={"worker2": "config"},
                ),
            ],
        ),
    ),
]


class TestHeadNodeConfig:
    def test_empty(self):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "__init__() missing 1 required positional argument: 'instance_type'"
            ),
        ):
            HeadNodeConfig()

    def test_invalid_instance_type(self):
        with pytest.raises(TypeError, match="'instance_type' must be a string"):
            HeadNodeConfig(instance_type=123)

    def test_invalid_resources(self):
        with pytest.raises(
            TypeError, match=re.escape("'resources' must be a Dict[str, float]")
        ):
            HeadNodeConfig(instance_type="m5.2xlarge", resources=["CPU", "GPU"])

        with pytest.raises(TypeError, match="'resources' values must be floats"):
            HeadNodeConfig(instance_type="m5.2xlarge", resources={"CPU": "123"})

        with pytest.raises(TypeError, match="'resources' keys must be strings"):
            HeadNodeConfig(instance_type="m5.2xlarge", resources={123: 1})

    def test_invalid_advanced_instance_config(self):
        with pytest.raises(
            TypeError,
            match=re.escape("'advanced_instance_config' must be a Dict[str, Any]"),
        ):
            HeadNodeConfig(
                instance_type="m5.2xlarge", advanced_instance_config=["foobar"]
            )

    def test_options(self):
        config = HeadNodeConfig(instance_type="m5.2xlarge")

        options = {
            "resources": {"CPU": 1},
            "advanced_instance_config": {"foo": 123},
        }

        # Test setting fields one at a time.
        for option, val in options.items():
            assert config.options(**{option: val}) == HeadNodeConfig(
                instance_type="m5.2xlarge", **{option: val}
            )

        # Test setting fields all at once.
        assert config.options(**options) == HeadNodeConfig(
            instance_type="m5.2xlarge", **options
        )


class TestWorkerNodeGroupConfig:
    def test_empty(self):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "__init__() missing 1 required positional argument: 'instance_type'"
            ),
        ):
            WorkerNodeGroupConfig()

    def test_invalid_instance_type(self):
        with pytest.raises(TypeError, match="'instance_type' must be a string"):
            WorkerNodeGroupConfig(instance_type=123)

    def test_market_type(self):
        for valid_market_type in list(MarketType):
            assert (
                WorkerNodeGroupConfig(
                    instance_type="m5.2xlarge", market_type=valid_market_type
                ).market_type
                == valid_market_type
            )
            assert (
                WorkerNodeGroupConfig(
                    instance_type="m5.2xlarge", market_type=str(valid_market_type)
                ).market_type
                == valid_market_type
            )

        with pytest.raises(
            ValueError, match="'DOES_NOT_EXIST' is not a valid MarketType"
        ):
            WorkerNodeGroupConfig(
                instance_type="m5.2xlarge", market_type="DOES_NOT_EXIST"
            )

    def test_name_defaults_to_instance_type(self):
        assert WorkerNodeGroupConfig(instance_type="m5.2xlarge").name == "m5.2xlarge"
        assert (
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", name="foobar").name
            == "foobar"
        )

    def test_invalid_name(self):
        with pytest.raises(TypeError, match="'name' must be a string"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", name=123)

        with pytest.raises(ValueError, match="'name' cannot be empty"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", name="")

    def test_invalid_min_nodes(self):
        WorkerNodeGroupConfig(instance_type="m5.2xlarge", min_nodes=0)
        with pytest.raises(TypeError, match="'min_nodes' must be an int"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", min_nodes="foobar")

        with pytest.raises(TypeError, match="'min_nodes' must be an int"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", min_nodes=1.5)

        with pytest.raises(ValueError, match="'min_nodes' must be >= 0"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", min_nodes=-1)

    def test_invalid_max_nodes(self):
        WorkerNodeGroupConfig(instance_type="m5.2xlarge", max_nodes=1)

        with pytest.raises(TypeError, match="'max_nodes' must be an int"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", max_nodes="foobar")

        with pytest.raises(TypeError, match="'max_nodes' must be an int"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", max_nodes=1.5)

        with pytest.raises(ValueError, match="'max_nodes' must be >= 1"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", max_nodes=0)

        with pytest.raises(
            ValueError, match=re.escape("'max_nodes' must be >= 'min_nodes' (2)")
        ):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", min_nodes=2, max_nodes=1)

    def test_invalid_resources(self):
        with pytest.raises(
            TypeError, match=re.escape("'resources' must be a Dict[str, float]")
        ):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", resources=["CPU", "GPU"])

        with pytest.raises(TypeError, match="'resources' values must be floats"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", resources={"CPU": "123"})

        with pytest.raises(TypeError, match="'resources' keys must be strings"):
            WorkerNodeGroupConfig(instance_type="m5.2xlarge", resources={123: 1})

    def test_invalid_advanced_instance_config(self):
        with pytest.raises(
            TypeError,
            match=re.escape("'advanced_instance_config' must be a Dict[str, Any]"),
        ):
            WorkerNodeGroupConfig(
                instance_type="m5.2xlarge", advanced_instance_config=["foobar"]
            )

    def test_options(self):
        config = WorkerNodeGroupConfig(instance_type="m5.2xlarge")

        options = {
            "name": "foobar",
            "resources": {"CPU": 1},
            "advanced_instance_config": {"foo": 123},
            "min_nodes": 5,
            "max_nodes": 500,
        }

        # Test setting fields one at a time.
        for option, val in options.items():
            assert config.options(**{option: val}) == WorkerNodeGroupConfig(
                instance_type="m5.2xlarge", **{option: val}
            )

        # Test setting fields all at once.
        assert config.options(**options) == WorkerNodeGroupConfig(
            instance_type="m5.2xlarge", **options
        )


class TestComputeConfig:
    def test_empty(self):
        ComputeConfig()

    def test_invalid_cloud(self):
        ComputeConfig(cloud="test-cloud")

        with pytest.raises(TypeError, match="'cloud' must be a string"):
            ComputeConfig(cloud=123)

    def test_invalid_zones(self):
        ComputeConfig(zones=["us-west-2a", "us-west-2b"])

        with pytest.raises(TypeError, match=re.escape("'zones' must be a List[str]")):
            ComputeConfig(zones={"hi"})

        with pytest.raises(TypeError, match=re.escape("'zones' must be a List[str]")):
            ComputeConfig(zones=["hi", 123])

        with pytest.raises(
            ValueError, match=re.escape("'zones' must not be an empty list")
        ):
            ComputeConfig(zones=[])

    def test_enable_cross_zone_scaling(self):
        assert ComputeConfig().enable_cross_zone_scaling is False
        assert (
            ComputeConfig(enable_cross_zone_scaling=False).enable_cross_zone_scaling
            is False
        )
        assert (
            ComputeConfig(enable_cross_zone_scaling=True).enable_cross_zone_scaling
            is True
        )

        with pytest.raises(
            TypeError, match="'enable_cross_zone_scaling' must be a boolean"
        ):
            ComputeConfig(enable_cross_zone_scaling="True")

    def test_invalid_advanced_instance_config(self):
        with pytest.raises(
            TypeError,
            match=re.escape("'advanced_instance_config' must be a Dict[str, Any]"),
        ):
            ComputeConfig(advanced_instance_config=["foobar"])

    def test_invalid_max_resources(self):
        with pytest.raises(
            TypeError, match=re.escape("'max_resources' must be a Dict[str, float]")
        ):
            ComputeConfig(max_resources=["CPU", "GPU"])

        with pytest.raises(TypeError, match="'max_resources' values must be floats"):
            ComputeConfig(max_resources={"CPU": "123"})

        with pytest.raises(
            ValueError,
            match="Only 'CPU' and 'GPU' are currently supported in 'max_resources', but got: {'custom'}",
        ):
            ComputeConfig(max_resources={"custom": 1})

    def test_invalid_head_node(self):
        ComputeConfig(head_node=HeadNodeConfig(instance_type="m5.2xlarge"))

        with pytest.raises(
            TypeError,
            match=re.escape(
                "'head_node' must be a HeadNodeConfig or corresponding dict"
            ),
        ):
            ComputeConfig(head_node=[123])

    def test_head_node_as_dict(self):
        assert ComputeConfig(
            head_node={"instance_type": "m5.2xlarge"}
        ).head_node == HeadNodeConfig(instance_type="m5.2xlarge")

    def test_invalid_worker_nodes(self):
        ComputeConfig(
            worker_nodes=[
                WorkerNodeGroupConfig(instance_type="m5.2xlarge"),
                WorkerNodeGroupConfig(instance_type="m5.4xlarge"),
            ]
        )

        with pytest.raises(
            TypeError,
            match=re.escape(
                "'worker_nodes' must be a list of WorkerNodeGroupConfigs or corresponding dicts"
            ),
        ):
            ComputeConfig(
                worker_nodes=[WorkerNodeGroupConfig(instance_type="m5.2xlarge"), 123]
            )

        with pytest.raises(
            TypeError,
            match=re.escape(
                "'worker_nodes' must be a list of WorkerNodeGroupConfigs or corresponding dicts"
            ),
        ):
            ComputeConfig(
                worker_nodes=[
                    WorkerNodeGroupConfig(instance_type="m5.2xlarge"),
                    HeadNodeConfig,
                ]
            )

    def test_worker_nodes_duplicate_names(self):
        with pytest.raises(
            ValueError,
            match="'worker_nodes' names must be unique, but got duplicate names: {'m5.2xlarge'}",
        ):
            ComputeConfig(
                worker_nodes=[
                    WorkerNodeGroupConfig(instance_type="m5.2xlarge"),
                    WorkerNodeGroupConfig(instance_type="m5.2xlarge"),
                ]
            )

        with pytest.raises(
            ValueError,
            match="'worker_nodes' names must be unique, but got duplicate names: {'foobar'}",
        ):
            ComputeConfig(
                worker_nodes=[
                    WorkerNodeGroupConfig(instance_type="m5.2xlarge", name="foobar"),
                    WorkerNodeGroupConfig(instance_type="m5.2xlarge", name="foobar"),
                ]
            )

        ComputeConfig(
            worker_nodes=[
                WorkerNodeGroupConfig(instance_type="m5.2xlarge"),
                WorkerNodeGroupConfig(instance_type="m5.2xlarge", name="foobar"),
            ]
        )

    def test_worker_nodes_as_dicts(self):
        assert ComputeConfig(
            worker_nodes=[
                {"instance_type": "m5.2xlarge"},
                {"instance_type": "m5.4xlarge"},
                WorkerNodeGroupConfig(instance_type="m5.8xlarge"),
            ]
        ).worker_nodes == [
            WorkerNodeGroupConfig(instance_type="m5.2xlarge"),
            WorkerNodeGroupConfig(instance_type="m5.4xlarge"),
            WorkerNodeGroupConfig(instance_type="m5.8xlarge"),
        ]

    @pytest.mark.parametrize("config_file", TEST_CONFIG_FILES)
    def test_from_config_file(self, config_file: ComputeConfigFile):
        if config_file.expected_error is not None:
            with pytest.raises(Exception, match=config_file.expected_error):
                ComputeConfig.from_yaml(config_file.get_path())

            return

        assert config_file.expected_config == ComputeConfig.from_yaml(
            config_file.get_path()
        )


class TestComputeConfigVersion:
    def test_missing_fields(self):
        with pytest.raises(TypeError, match="missing 3 required positional arguments"):
            ComputeConfigVersion()

        with pytest.raises(TypeError, match="missing 2 required positional arguments"):
            ComputeConfigVersion(name="foobar")

        with pytest.raises(TypeError, match="missing 1 required positional argument"):
            ComputeConfigVersion(name="foobar", id="barbaz")

    def test_must_have_version_tag(self):
        with pytest.raises(
            ValueError, match="'name' must be in the format: '<name>:<version>'"
        ):
            ComputeConfigVersion(
                name="test-name-no-version", id="test-id", config=ComputeConfig(),
            )

    def test_basic(self):
        ComputeConfigVersion(
            name="test-name:1", id="test-id", config=ComputeConfig(),
        )



from typing import Optional

import click

from anyscale.controllers.fine_tune_controller import FineTuneController
from anyscale.models.fine_tune_model import FineTuneConfig


@click.group(
    "fine-tuning", help="Interact with fine-tuning jobs running on Anyscale.",
)
def fine_tune_cli() -> None:
    pass


@fine_tune_cli.command(name="submit")
@click.argument("base-model", required=True)
@click.option("--train-file", required=True, help="The path of the training file.")
@click.option(
    "--valid-file",
    required=False,
    default=None,
    help="The path of the validation file.",
)
@click.option("--cloud-id", required=True, help="The id of the cloud")
@click.option(
    "--suffix",
    required=False,
    default=None,
    help="The suffix of the fine-tuned model.",
)
@click.option(
    "--version",
    required=False,
    default=None,
    help=(
        "The version of ray-llm expected to be used for the fine-tuned model. "
        "If not specified, the latest version will be used."
    ),
)
@click.option(
    "--instance-type",
    required=False,
    default=None,
    help=(
        "The instance type to use for the fine-tuning job. "
        "Please refer to docs at https://docs.anyscale.com/endpoints/fine-tuning/get-started for available instance types."
        "If not specified, the default instance type will be used."
    ),
)
def submit(
    base_model: str,
    train_file: str,
    valid_file: Optional[str],
    cloud_id: str,
    suffix: Optional[str],
    version: Optional[str],
    instance_type: Optional[str],
) -> None:
    """
    Submits a fine-tuning job.

    Example usage:

        anyscale fine-tuning submit --model meta-llama/Llama-2-7b-chat-hf --train-file train.jsonl --cloud-id CLOUD_ID
    """
    fine_tune_controller = FineTuneController()
    config = FineTuneConfig(
        base_model=base_model,
        train_file=train_file,
        valid_file=valid_file,
        cloud_id=cloud_id,
        suffix=suffix,
        version=version,
        instance_type=instance_type,
    )
    fine_tune_controller.fine_tune(config=config)



import ipaddress
from typing import Set, Union

import pytest

from anyscale.utils.network_verification import (
    check_inbound_firewall_permissions,
    Direction,
    FirewallRule,
    Protocol,
)


@pytest.mark.parametrize(
    ("ports_in", "output"),
    [
        pytest.param(None, set(range(1, 65536)), id="None"),
        pytest.param({}, set(range(1, 65536)), id="EmptySet"),
        ({22, 443}, {22, 443}),
        ({"22-23", 443}, {22, 23, 443}),
        ({"22-23", 443, "440-443"}, {22, 23, 440, 441, 442, 443}),
        ({"1", "300"}, {1, 300}),
    ],
)
def test_firewall_rule_port_creation(ports_in: Set[Union[str, int]], output: Set[int]):
    assert (
        FirewallRule(
            protocol=Protocol.tcp,
            direction=Direction.INGRESS,
            ports=ports_in,  # type: ignore
            network=ipaddress.IPv4Network("127.0.0.0/24"),
        ).ports
        == output
    )


@pytest.mark.parametrize(
    ("protos_in", "output"),
    [
        ("tcp", "tcp"),
        ("all", "all"),
        (17, "udp"),
        ("17", "udp"),
        ("1024", "other"),
        ("gre", "other"),
    ],
)
def test_firewall_rule_protocol_resolution(protos_in: Union[str, int], output: str):
    assert (
        FirewallRule(
            protocol=Protocol.from_val(protos_in),
            direction=Direction.INGRESS,
            ports=None,
            network=ipaddress.IPv4Network("127.0.0.0/24"),
        ).protocol.value
        == output
    )


@pytest.mark.parametrize("direction", [Direction.INGRESS, Direction.EGRESS])
@pytest.mark.parametrize(
    ("firewall_ports", "wanted_ports", "result"),
    [
        pytest.param({22}, {22}, True, id="simple"),
        pytest.param(None, None, True, id="all"),
        pytest.param({22, 44}, None, False, id="MissingPorts"),
        pytest.param(["1-65535"], None, True, id="MissingZero"),
        pytest.param(
            ["1-65535"], set(range(65536)), True, id="AccidentallyWantPortZero"
        ),
    ],
)
def test_check_inbound_firewall_permissions_ports(
    firewall_ports, wanted_ports, result: bool, direction: Direction
):
    check = check_inbound_firewall_permissions(
        [
            FirewallRule(
                direction=direction,
                protocol=Protocol.tcp,
                network=ipaddress.IPv4Network("127.0.0.0/24"),
                ports=firewall_ports,
            )
        ],
        Protocol.tcp,
        wanted_ports,
        ipaddress.IPv4Network("127.0.0.0/24"),
    )
    if direction == Direction.INGRESS:
        assert check == result
    else:
        assert not check


def test_check_inbound_firewall_permissions_ports_over_several_rules():
    assert check_inbound_firewall_permissions(
        [
            FirewallRule(
                direction="INGRESS",
                protocol=Protocol.tcp,
                network=ipaddress.IPv4Network("127.0.0.0/24"),
                ports=ports,
            )
            for ports in [["20-30"], [443, 445], [2000]]
        ],
        Protocol.tcp,
        {22, 443, 2000},
        ipaddress.IPv4Network("127.0.0.0/24"),
    )


def test_check_inbound_firewall_permissions_ports_without_checking_source_range():
    assert check_inbound_firewall_permissions(
        [
            FirewallRule(
                direction="INGRESS",
                protocol=Protocol.tcp,
                network=ipaddress.IPv4Network("127.0.0.0/24"),
                ports=[22, 443],
            )
        ],
        Protocol.tcp,
        {22, 443},
        ipaddress.IPv4Network("0.0.0.0/0"),
        False,
    )


@pytest.mark.parametrize(
    ("firewall_proto", "wanted_proto", "result"),
    [
        pytest.param("all", "all", True, id="all"),
        pytest.param("all", "tcp", True, id="Overlybroad"),
        pytest.param("all", "icmp", True, id="ICMP"),
        pytest.param("tcp", "all", False, id="insufficient"),
        pytest.param("tcp", "udp", False, id="Mismatch"),
    ],
)
def test_check_inbound_firewall_permissions_protocols(
    firewall_proto: str, wanted_proto: str, result: bool
):
    assert (
        check_inbound_firewall_permissions(
            [
                FirewallRule(
                    direction=Direction.INGRESS,
                    protocol=Protocol.from_val(firewall_proto),
                    network=ipaddress.IPv4Network("127.0.0.0/24"),
                    ports=None,
                )
            ],
            Protocol.from_val(wanted_proto),
            None,
            ipaddress.IPv4Network("127.0.0.0/24"),
        )
        == result
    )


@pytest.mark.parametrize(
    ("firewall_net", "wanted_net", "result"),
    [
        pytest.param("10.0.0.0/8", "10.0.0.0/20", True, id="larger"),
        pytest.param("10.0.0.0/8", "10.0.0.0/8", True, id="exactMatch"),
        pytest.param("0.0.0.0/0", "172.23.0.0/16", True, id="AllFirewall"),
        pytest.param("0.0.0.0/0", "0.0.0.0/0", True, id="AllAll"),
        pytest.param("10.0.0.0/9", "0.0.0.0/0", False, id="WantAll"),
        pytest.param("::/0", "0.0.0.0/0", False, id="V6Firewall"),
        pytest.param("0.0.0.0/0", "::/0", False, id="V6Source"),
        pytest.param("10.0.0.0/10", "10.0.0.0/8", False, id="FirewallSmall"),
    ],
)
def test_check_inbound_firewall_permissions_networks(
    firewall_net: str, wanted_net: str, result: bool
):
    assert (
        check_inbound_firewall_permissions(
            [
                FirewallRule(
                    direction=Direction.INGRESS,
                    protocol=Protocol.all,
                    network=ipaddress.ip_network(firewall_net),
                    ports=None,
                )
            ],
            Protocol.all,
            None,
            ipaddress.ip_network(wanted_net),
        )
        == result
    )



from typing import Optional
from unittest.mock import call, Mock

from openapi_client.rest import ApiException
import pytest

from anyscale.client.openapi_client.models import Cloud, CloudResponse
from anyscale.cloud import (
    get_all_clouds,
    get_cloud_id_and_name,
    get_cloud_json_from_id,
    get_last_used_cloud,
    get_organization_default_cloud,
)
from anyscale.sdk.anyscale_client.models import ListResponseMetadata
from anyscale.sdk.anyscale_client.models.cloud_list_response import CloudListResponse


def test_get_cloud_json_from_id(cloud_test_data: Cloud) -> None:
    mock_api_client = Mock()
    mock_api_client.get_cloud_api_v2_clouds_cloud_id_get.return_value = CloudResponse(
        result=cloud_test_data
    )
    cloud_json = get_cloud_json_from_id(cloud_test_data.id, api_client=mock_api_client)
    expected_json = {
        "id": cloud_test_data.id,
        "name": cloud_test_data.name,
        "provider": cloud_test_data.provider,
        "region": cloud_test_data.region,
        "credentials": cloud_test_data.credentials,
        "config": cloud_test_data.config,
        "state": "ACTIVE",
    }
    assert cloud_json == expected_json
    mock_api_client.get_cloud_api_v2_clouds_cloud_id_get.assert_called_once_with(
        cloud_id=cloud_test_data.id
    )


def test_get_cloud_json_from_id_api_error(base_mock_api_client: Mock) -> None:
    base_mock_api_client.get_cloud_api_v2_clouds_cloud_id_get = Mock(
        side_effect=ApiException()
    )
    cloud_json = get_cloud_json_from_id("cld_1234", base_mock_api_client)
    expected_json = {
        "error": {
            "cloud_id": "cld_1234",
            "message": "The cloud with id, cld_1234 has been deleted. Please create a new cloud with `anyscale cloud setup`.",
        }
    }
    assert cloud_json == expected_json


def test_get_cloud_id_and_name_no_args(
    base_mock_api_client: Mock, cloud_test_data: Cloud
) -> None:
    with pytest.raises(Exception):  # noqa: PT011
        get_cloud_id_and_name(base_mock_api_client)


def test_get_cloud_id_and_name_two_args(
    base_mock_api_client: Mock, cloud_test_data: Cloud
) -> None:
    with pytest.raises(Exception):  # noqa: PT011
        get_cloud_id_and_name(
            base_mock_api_client, cloud_test_data.id, cloud_test_data.name
        )


def test_get_cloud_id_and_name_id_only(
    base_mock_api_client: Mock, cloud_test_data: Cloud
) -> None:
    base_mock_api_client.get_cloud_api_v2_clouds_cloud_id_get = Mock(
        return_value=CloudResponse(result=cloud_test_data)
    )
    result_id, result_name = get_cloud_id_and_name(
        base_mock_api_client, cloud_id=cloud_test_data.id
    )
    assert result_id == cloud_test_data.id
    assert result_name == cloud_test_data.name
    base_mock_api_client.get_cloud_api_v2_clouds_cloud_id_get.assert_called_once()


def test_get_cloud_id_and_name_name_only(
    base_mock_api_client: Mock, cloud_test_data: Cloud
) -> None:
    base_mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post = Mock(
        return_value=CloudResponse(result=cloud_test_data)
    )
    result_id, result_name = get_cloud_id_and_name(
        base_mock_api_client, cloud_name=cloud_test_data.name
    )
    assert result_id == cloud_test_data.id
    assert result_name == cloud_test_data.name
    base_mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post.assert_called_once()


@pytest.mark.parametrize("default_cloud_exists", [True, False])
@pytest.mark.parametrize("default_cloud_permissions_exist", [True, False])
def test_get_organization_default_cloud(
    default_cloud_exists: bool, default_cloud_permissions_exist: bool,
):
    mock_api_client = Mock()
    mock_user = Mock(
        organizations=[
            Mock(
                default_cloud_id="test_default_cloud_id"
                if default_cloud_exists
                else None
            )
        ]
    )
    mock_api_client.get_user_info_api_v2_userinfo_get = Mock(
        return_value=Mock(result=mock_user)
    )
    mock_cloud = Mock(id="test_default_cloud_id")
    mock_cloud.name = "test_default_cloud_name"
    if not default_cloud_permissions_exist:
        mock_api_client.get_cloud_api_v2_clouds_cloud_id_get = Mock(
            return_value=Mock(result=None)
        )
    else:
        mock_api_client.get_cloud_api_v2_clouds_cloud_id_get = Mock(
            return_value=Mock(result=mock_cloud)
        )

    if not default_cloud_exists or not default_cloud_permissions_exist:
        assert get_organization_default_cloud(mock_api_client) is None
    else:
        assert (
            get_organization_default_cloud(mock_api_client) == "test_default_cloud_name"
        )
    mock_api_client.get_user_info_api_v2_userinfo_get.assert_called_once_with()
    if default_cloud_exists:
        mock_api_client.get_cloud_api_v2_clouds_cloud_id_get.assert_called_once_with(
            cloud_id="test_default_cloud_id"
        )


@pytest.mark.parametrize("cloud_id", ["test_cloud_id", None])
def test_get_last_used_cloud(cloud_id: Optional[str]):
    mock_anyscale_api_client = Mock()
    mock_anyscale_api_client.get_project = Mock(
        return_value=Mock(result=Mock(last_used_cloud_id=cloud_id))
    )
    mock_cloud = Mock(id="test_cloud_id")
    mock_cloud.name = "test_cloud_name"
    mock_anyscale_api_client.get_cloud = Mock(return_value=Mock(result=mock_cloud))
    mock_anyscale_api_client.search_clouds = Mock(
        return_value=Mock(results=[mock_cloud], metadata=Mock(next_paging_token=None))
    )

    assert (
        get_last_used_cloud("test_project_id", mock_anyscale_api_client)
        == "test_cloud_name"
    )
    mock_anyscale_api_client.get_project.assert_called_once_with("test_project_id")
    if cloud_id:
        mock_anyscale_api_client.get_cloud.assert_called_once_with("test_cloud_id")
    else:
        mock_anyscale_api_client.search_clouds.assert_called_once_with(
            {"paging": {"count": 50}}
        )


def test_get_all_clouds() -> None:
    mock_cloud_1 = Mock()
    mock_cloud_1.name = "cloud_1"
    mock_cloud_2 = Mock()
    mock_cloud_2.name = "cloud_2"

    mock_anyscale_api_client = Mock()
    mock_anyscale_api_client.search_clouds.side_effect = [
        CloudListResponse(
            results=[mock_cloud_1],
            metadata=ListResponseMetadata(
                total=2, next_paging_token="next_paging_token"
            ),
        ),
        CloudListResponse(
            results=[mock_cloud_2], metadata=ListResponseMetadata(total=2),
        ),
    ]

    all_clouds = get_all_clouds(mock_anyscale_api_client)
    assert all_clouds == [mock_cloud_1, mock_cloud_2]
    mock_anyscale_api_client.search_clouds.assert_has_calls(
        [
            call({"paging": {"count": 50}}),
            call({"paging": {"count": 50, "paging_token": "next_paging_token"}}),
        ]
    )



import click

from anyscale.commands.llm.dataset_commands import dataset_cli


@click.group(
    "llm", help="Interact with Anyscale's LLM APIs.", hidden=True,
)
def llm_cli():
    pass


llm_cli.add_command(dataset_cli)
